#
# Copyright 2022 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Tests for ceilometer/publisher/tcp.py"""

from unittest import mock

import msgpack
from oslo_utils import netutils
from oslo_utils import timeutils
from oslotest import base

from ceilometer.publisher.tcp import TCPPublisher
from ceilometer.publisher import utils
from ceilometer import sample
from ceilometer import service


COUNTER_SOURCE = 'testsource'


class TestTCPPublisher(base.BaseTestCase):
    test_data = [
        sample.Sample(
            name='test',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id='test_run_tasks',
            timestamp=timeutils.utcnow().isoformat(),
            resource_metadata={'name': 'TestPublish'},
            source=COUNTER_SOURCE,
        ),
        sample.Sample(
            name='test',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id='test_run_tasks',
            timestamp=timeutils.utcnow().isoformat(),
            resource_metadata={'name': 'TestPublish'},
            source=COUNTER_SOURCE,
        ),
        sample.Sample(
            name='test2',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id='test_run_tasks',
            timestamp=timeutils.utcnow().isoformat(),
            resource_metadata={'name': 'TestPublish'},
            source=COUNTER_SOURCE,
        ),
        sample.Sample(
            name='test2',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id='test_run_tasks',
            timestamp=timeutils.utcnow().isoformat(),
            resource_metadata={'name': 'TestPublish'},
            source=COUNTER_SOURCE,
        ),
        sample.Sample(
            name='test3',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id='test_run_tasks',
            timestamp=timeutils.utcnow().isoformat(),
            resource_metadata={'name': 'TestPublish'},
            source=COUNTER_SOURCE,
        ),
    ]

    @staticmethod
    def _make_fake_socket(published):
        def _fake_socket_create_connection(inet_addr):
            def record_data(msg):
                msg_length = int.from_bytes(msg[0:8], "little")
                published.append(msg[8:msg_length + 8])

            tcp_socket = mock.Mock()
            tcp_socket.send = record_data
            return tcp_socket

        return _fake_socket_create_connection

    def setUp(self):
        super().setUp()
        self.CONF = service.prepare_service([], [])
        self.CONF.publisher.telemetry_secret = 'not-so-secret'

    def test_published(self):
        self.data_sent = []
        with mock.patch('ceilometer.publisher.tcp.socket.create_connection',
                        self._make_fake_socket(self.data_sent)):
            publisher = TCPPublisher(self.CONF,
                                     netutils.urlsplit('tcp://somehost'))
        publisher.publish_samples(self.test_data)

        self.assertEqual(5, len(self.data_sent))

        sent_counters = []

        for data in self.data_sent:
            counter = msgpack.loads(data, raw=False)
            sent_counters.append(counter)

        # Check that counters are equal
        def sort_func(counter):
            return counter['counter_name']

        counters = [utils.meter_message_from_counter(d,
                                                     "not-so-secret",
                                                     publisher.conf.host)
                    for d in self.test_data]
        counters.sort(key=sort_func)
        sent_counters.sort(key=sort_func)
        self.assertEqual(counters, sent_counters)

    def _make_disconnecting_socket(self):
        def _fake_socket_create_connection(inet_addr):
            def record_data(msg):
                if not self.connections:
                    self.connections = True
                    raise OSError
                msg_length = int.from_bytes(msg[0:8], "little")
                self.data_sent.append(msg[8:msg_length + 8])

            tcp_socket = mock.MagicMock()
            tcp_socket.send = record_data
            return tcp_socket

        return _fake_socket_create_connection

    def test_reconnect(self):
        self.data_sent = []
        self.connections = False
        with mock.patch('ceilometer.publisher.tcp.socket.create_connection',
                        self._make_disconnecting_socket()):
            publisher = TCPPublisher(self.CONF,
                                     netutils.urlsplit('tcp://somehost'))
            publisher.publish_samples(self.test_data)

        sent_counters = []

        for data in self.data_sent:
            counter = msgpack.loads(data, raw=False)
            sent_counters.append(counter)

        # Check that counters are equal
        def sort_func(counter):
            return counter['counter_name']

        counters = [utils.meter_message_from_counter(d,
                                                     "not-so-secret",
                                                     publisher.conf.host)
                    for d in self.test_data]
        counters.sort(key=sort_func)
        sent_counters.sort(key=sort_func)
        self.assertEqual(counters, sent_counters)

    @staticmethod
    def _raise_OSError(*args):
        raise OSError

    def _make_broken_socket(self, inet_addr):
        tcp_socket = mock.Mock()
        tcp_socket.send = self._raise_OSError
        return tcp_socket

    def test_publish_error(self):
        with mock.patch('ceilometer.publisher.tcp.socket.create_connection',
                        self._make_broken_socket):
            publisher = TCPPublisher(self.CONF,
                                     netutils.urlsplit('tcp://localhost'))
            publisher.publish_samples(self.test_data)



#
# Copyright 2022 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Publish a sample using a TCP mechanism
"""

import socket

import msgpack
from oslo_log import log
from oslo_utils import netutils

import ceilometer
from ceilometer import publisher
from ceilometer.publisher import utils

LOG = log.getLogger(__name__)


class TCPPublisher(publisher.ConfigPublisherBase):
    def __init__(self, conf, parsed_url):
        super().__init__(conf, parsed_url)
        self.inet_addr = netutils.parse_host_port(
            parsed_url.netloc, default_port=4952)
        self.socket = None
        self.connect_socket()

    def connect_socket(self):
        try:
            self.socket = socket.create_connection(self.inet_addr)
            return True
        except socket.gaierror:
            LOG.error("Unable to resolv the remote %(host)s",
                      {'host': self.inet_addr[0],
                       'port': self.inet_addr[1]})
        except TimeoutError:
            LOG.error("Unable to connect to the remote endpoint "
                      "%(host)s:%(port)d. The connection timed out.",
                      {'host': self.inet_addr[0],
                       'port': self.inet_addr[1]})
        except ConnectionRefusedError:
            LOG.error("Unable to connect to the remote endpoint "
                      "%(host)s:%(port)d. Connection refused.",
                      {'host': self.inet_addr[0],
                       'port': self.inet_addr[1]})
        return False

    def publish_samples(self, samples):
        """Send a metering message for publishing

        :param samples: Samples from pipeline after transformation
        """

        for sample in samples:
            msg = utils.meter_message_from_counter(
                sample, self.conf.publisher.telemetry_secret, self.conf.host)
            LOG.debug("Publishing sample %(msg)s over TCP to "
                      "%(host)s:%(port)d",
                      {'msg': msg,
                       'host': self.inet_addr[0],
                       'port': self.inet_addr[1]})
            encoded_msg = msgpack.dumps(msg, use_bin_type=True)
            msg_len = len(encoded_msg).to_bytes(8, 'little')
            if self.socket:
                try:
                    self.socket.send(msg_len + encoded_msg)
                    continue
                except OSError:
                    LOG.warning("Unable to send sample over TCP, trying "
                                "to reconnect and resend the message")
            if self.connect_socket():
                try:
                    self.socket.send(msg_len + encoded_msg)
                    continue
                except OSError:
                    pass
            LOG.error("Unable to reconnect and resend sample over TCP")
            # NOTE (jokke): We do not handle exceptions in the calling code
            # so raising the exception from here needs quite a bit more work.
            # Same time we don't want to spam the retry messages as it's
            # unlikely to change between iterations on this loop. 'break'
            # rather than 'return' even the end result is the same feels
            # more appropriate for now.
            break

    def publish_events(self, events):
        """Send an event message for publishing

        :param events: events from pipeline after transformation
        """
        raise ceilometer.NotImplementedError



#
# Copyright 2016 Mirantis, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import fixtures

from ceilometer import declarative
from ceilometer.tests import base


class TestDefinition(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.configs = [
            "_field1",
            "_field2|_field3",
            {'fields': 'field4.`split(., 1, 1)`'},
            {'fields': ['field5.arg', 'field6'], 'type': 'text'}
        ]
        self.parser = mock.MagicMock()
        parser_patch = fixtures.MockPatch(
            "jsonpath_rw_ext.parser.ExtentedJsonPathParser.parse",
            new=self.parser)
        self.useFixture(parser_patch)

    def test_caching_parsers(self):
        for config in self.configs * 2:
            declarative.Definition("test", config, mock.MagicMock())
        self.assertEqual(4, self.parser.call_count)
        self.parser.assert_has_calls([
            mock.call("_field1"),
            mock.call("_field2|_field3"),
            mock.call("field4.`split(., 1, 1)`"),
            mock.call("(field5.arg)|(field6)"),
        ])



#
# Copyright 2013-2014 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from ceilometer import middleware
from ceilometer import service
from ceilometer.tests import base


HTTP_REQUEST = {
    'ctxt': {'auth_token': '3d8b13de1b7d499587dfc69b77dc09c2',
             'is_admin': True,
             'project_id': '7c150a59fe714e6f9263774af9688f0e',
             'quota_class': None,
             'read_deleted': 'no',
             'remote_address': '10.0.2.15',
             'request_id': 'req-d68b36e0-9233-467f-9afb-d81435d64d66',
             'roles': ['admin'],
             'timestamp': '2012-05-08T20:23:41.425105',
             'user_id': '1e3ce043029547f1a61c1996d1a531a2'},
    'event_type': 'http.request',
    'payload': {'request': {'HTTP_X_FOOBAR': 'foobaz',
                            'HTTP_X_USER_ID': 'jd-x32',
                            'HTTP_X_PROJECT_ID': 'project-id',
                            'HTTP_X_SERVICE_NAME': 'nova'}},
    'priority': 'INFO',
    'publisher_id': 'compute.vagrant-precise',
    'metadata': {'message_id': 'dae6f69c-00e0-41c0-b371-41ec3b7f4451',
                 'timestamp': '2012-05-08 20:23:48.028195'},
}

HTTP_RESPONSE = {
    'ctxt': {'auth_token': '3d8b13de1b7d499587dfc69b77dc09c2',
             'is_admin': True,
             'project_id': '7c150a59fe714e6f9263774af9688f0e',
             'quota_class': None,
             'read_deleted': 'no',
             'remote_address': '10.0.2.15',
             'request_id': 'req-d68b36e0-9233-467f-9afb-d81435d64d66',
             'roles': ['admin'],
             'timestamp': '2012-05-08T20:23:41.425105',
             'user_id': '1e3ce043029547f1a61c1996d1a531a2'},
    'event_type': 'http.response',
    'payload': {'request': {'HTTP_X_FOOBAR': 'foobaz',
                            'HTTP_X_USER_ID': 'jd-x32',
                            'HTTP_X_PROJECT_ID': 'project-id',
                            'HTTP_X_SERVICE_NAME': 'nova'},
                'response': {'status': '200 OK'}},
    'priority': 'INFO',
    'publisher_id': 'compute.vagrant-precise',
    'metadata': {'message_id': 'dae6f69c-00e0-41c0-b371-41ec3b7f4451',
                 'timestamp': '2012-05-08 20:23:48.028195'},
}


class TestNotifications(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.CONF = service.prepare_service([], [])
        self.setup_messaging(self.CONF)

    def test_process_request_notification(self):
        sample = list(middleware.HTTPRequest(
            mock.Mock(), mock.Mock()).build_sample(HTTP_REQUEST))[0]
        self.assertEqual(HTTP_REQUEST['payload']['request']['HTTP_X_USER_ID'],
                         sample.user_id)
        self.assertEqual(HTTP_REQUEST['payload']['request']
                         ['HTTP_X_PROJECT_ID'], sample.project_id)
        self.assertEqual(HTTP_REQUEST['payload']['request']
                         ['HTTP_X_SERVICE_NAME'], sample.resource_id)
        self.assertEqual(1, sample.volume)

    def test_process_response_notification(self):
        sample = list(middleware.HTTPResponse(
            mock.Mock(), mock.Mock()).build_sample(HTTP_RESPONSE))[0]
        self.assertEqual(HTTP_RESPONSE['payload']['request']['HTTP_X_USER_ID'],
                         sample.user_id)
        self.assertEqual(HTTP_RESPONSE['payload']['request']
                         ['HTTP_X_PROJECT_ID'], sample.project_id)
        self.assertEqual(HTTP_RESPONSE['payload']['request']
                         ['HTTP_X_SERVICE_NAME'], sample.resource_id)
        self.assertEqual(1, sample.volume)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import fixtures
from openstack.dns.v2 import recordset
from openstack.dns.v2 import zone
from oslotest import base

from ceilometer.dns import designate
from ceilometer.dns import discovery
from ceilometer.polling import manager
from ceilometer.polling import plugin_base
from ceilometer import service


class _BaseTestDNSPollster(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.addCleanup(mock.patch.stopall)
        self.CONF = service.prepare_service([], [])
        # Mock the openstack.connection.Connection to avoid auth issues
        with mock.patch('openstack.connection.Connection'):
            self.manager = manager.AgentManager(0, self.CONF)
        plugin_base._get_keystone = mock.Mock()
        catalog = (plugin_base._get_keystone.session.auth.get_access.
                   return_value.service_catalog)
        catalog.get_endpoints = mock.MagicMock(
            return_value={'dns': mock.ANY})

    @staticmethod
    def fake_get_zones():
        return [
            zone.Zone(
                connection=None,
                id='zone-1-uuid',
                name='example.com.',
                email='admin@example.com',
                ttl=3600,
                description='Example zone',
                type='PRIMARY',
                status='ACTIVE',
                action='NONE',
                serial=1234567890,
                pool_id='pool-1',
                project_id='tenant-1-uuid',
            ),
            zone.Zone(
                connection=None,
                id='zone-2-uuid',
                name='test.org.',
                email='admin@test.org',
                ttl=7200,
                description='Test zone',
                type='PRIMARY',
                status='PENDING',
                action='CREATE',
                serial=1234567891,
                pool_id='pool-1',
                project_id='tenant-2-uuid',
            ),
            zone.Zone(
                connection=None,
                id='zone-3-uuid',
                name='error.net.',
                email='admin@error.net',
                ttl=1800,
                description='Error zone',
                type='PRIMARY',
                status='ERROR',
                action='UPDATE',
                serial=1234567892,
                pool_id='pool-2',
                project_id='tenant-1-uuid',
            ),
        ]

    @staticmethod
    def fake_get_recordsets():
        return [
            recordset.Recordset(
                connection=None,
                id='rs-1-uuid',
                name='www.example.com.',
                type='A',
                records=['192.168.1.1'],
                ttl=3600,
            ),
            recordset.Recordset(
                connection=None,
                id='rs-2-uuid',
                name='mail.example.com.',
                type='MX',
                records=['10 mail.example.com.'],
                ttl=3600,
            ),
        ]


class TestZoneStatusPollster(_BaseTestDNSPollster):

    def setUp(self):
        super().setUp()
        self.pollster = designate.ZoneStatusPollster(self.CONF)
        fake_zones = self.fake_get_zones()
        self.useFixture(fixtures.MockPatch(
            'ceilometer.designate_client.Client.zones_list',
            return_value=fake_zones))

    def test_zone_get_samples(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_zones()))

        self.assertEqual(len(self.fake_get_zones()), len(samples))
        for field in self.pollster.FIELDS:
            self.assertEqual(
                getattr(self.fake_get_zones()[0], field),
                samples[0].resource_metadata[field])

    def test_zone_status_volume(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_zones()))

        # ACTIVE = 1, PENDING = 2, ERROR = 3
        self.assertEqual(1, samples[0].volume)
        self.assertEqual(2, samples[1].volume)
        self.assertEqual(3, samples[2].volume)

    def test_get_zone_meter_names(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_zones()))
        self.assertEqual({'dns.zone.status'},
                         {s.name for s in samples})

    def test_zone_discovery(self):
        with mock.patch('openstack.connection.Connection'):
            discovered_zones = discovery.ZoneDiscovery(
                self.CONF).discover(self.manager)
        self.assertEqual(len(self.fake_get_zones()),
                         len(list(discovered_zones)))


class TestZoneRecordsetCountPollster(_BaseTestDNSPollster):

    def setUp(self):
        super().setUp()
        with mock.patch('openstack.connection.Connection'):
            self.pollster = designate.ZoneRecordsetCountPollster(self.CONF)
        self.useFixture(fixtures.MockPatch(
            'ceilometer.designate_client.Client.recordsets_list',
            return_value=self.fake_get_recordsets()))

    def test_zone_recordsets_volume(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_zones()))

        # Each zone should have the mocked number of recordsets
        num_recordsets = len(self.fake_get_recordsets())
        for sample in samples:
            self.assertEqual(num_recordsets, sample.volume)

    def test_get_zone_meter_names(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_zones()))
        self.assertEqual({'dns.zone.recordsets'},
                         {s.name for s in samples})


class TestZoneTTLPollster(_BaseTestDNSPollster):

    def setUp(self):
        super().setUp()
        self.pollster = designate.ZoneTTLPollster(self.CONF)

    def test_zone_ttl_volume(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_zones()))
        self.assertEqual(3600, samples[0].volume)
        self.assertEqual(7200, samples[1].volume)
        self.assertEqual(1800, samples[2].volume)

    def test_get_zone_meter_names(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_zones()))
        self.assertEqual({'dns.zone.ttl'},
                         {s.name for s in samples})


class TestZoneSerialPollster(_BaseTestDNSPollster):

    def setUp(self):
        super().setUp()
        self.pollster = designate.ZoneSerialPollster(self.CONF)

    def test_zone_serial_volume(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_zones()))
        self.assertEqual(1234567890, samples[0].volume)
        self.assertEqual(1234567891, samples[1].volume)
        self.assertEqual(1234567892, samples[2].volume)

    def test_get_zone_meter_names(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_zones()))
        self.assertEqual({'dns.zone.serial'},
                         {s.name for s in samples})


class TestZonePollsterUnknownStatus(_BaseTestDNSPollster):

    def test_unknown_zone_status(self):
        pollster = designate.ZoneStatusPollster(self.CONF)
        fake_zone = zone.Zone(
            connection=None,
            id='zone-unknown-uuid',
            name='unknown.com.',
            email='admin@unknown.com',
            ttl=3600,
            description='Unknown zone',
            type='PRIMARY',
            status='UNKNOWN_STATUS',
            action='NONE',
            serial=1234567893,
            pool_id='pool-1',
            project_id='tenant-1-uuid',
        )
        samples = list(pollster.get_samples(
            self.manager, {}, resources=[fake_zone]))
        self.assertEqual(1, len(samples))
        self.assertEqual(-1, samples[0].volume)



# Copyright 2014 Intel Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ceilometer.ipmi.platform import exception as nmexcept
from ceilometer.tests.unit.ipmi.platform import ipmitool_test_data as test_data


def get_sensor_status_init(parameter=''):
    return (' 01\n', '')


def get_sensor_status_uninit(parameter=''):
    return (' 00\n', '')


def init_sensor_agent(parameter=''):
    return (' 00\n', '')


def execute(*cmd, **kwargs):

    data = {
        test_data.sdr_info_cmd: test_data.sdr_info,
        test_data.read_sensor_temperature_cmd: test_data.sensor_temperature,
        test_data.read_sensor_voltage_cmd: test_data.sensor_voltage,
        test_data.read_sensor_current_cmd: test_data.sensor_current,
        test_data.read_sensor_fan_cmd: test_data.sensor_fan,
    }

    cmd_str = "".join(cmd)
    return data[cmd_str]


def execute_without_ipmi(*cmd, **kwargs):
    raise nmexcept.IPMIException



#
# Copyright 2013 Rackspace Hosting.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fnmatch
import os

from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from ceilometer import declarative
from ceilometer.event import models
from ceilometer.i18n import _

OPTS = [
    cfg.StrOpt('definitions_cfg_file',
               default="event_definitions.yaml",
               help="Configuration file for event definitions."
               ),
    cfg.BoolOpt('drop_unmatched_notifications',
                default=False,
                help='Drop notifications if no event definition matches. '
                '(Otherwise, we convert them with just the default traits)'),
    cfg.MultiStrOpt('store_raw',
                    default=[],
                    help='Store the raw notification for select priority '
                    'levels (info and/or error). By default, raw details are '
                    'not captured.')
]

LOG = log.getLogger(__name__)


class TraitDefinition(declarative.Definition):
    def __init__(self, name, trait_cfg, plugin_manager):
        super().__init__(name, trait_cfg, plugin_manager)
        type_name = (trait_cfg.get('type', 'text')
                     if isinstance(trait_cfg, dict) else 'text')
        self.trait_type = models.Trait.get_type_by_name(type_name)
        if self.trait_type is None:
            raise declarative.EventDefinitionException(
                _("Invalid trait type '%(type)s' for trait %(trait)s")
                % dict(type=type_name, trait=name), self.cfg)

    def to_trait(self, notification_body):
        value = self.parse(notification_body)
        if value is None:
            return None

        # NOTE(mdragon): some openstack projects (mostly Nova) emit ''
        # for null fields for things like dates.
        if self.trait_type != models.Trait.TEXT_TYPE and value == '':
            return None

        value = models.Trait.convert_value(self.trait_type, value)
        return models.Trait(self.name, self.trait_type, value)


class EventDefinition:

    DEFAULT_TRAITS = dict(
        service=dict(type='text', fields='publisher_id'),
        request_id=dict(type='text', fields='ctxt.request_id'),
        project_id=dict(type='text', fields=['payload.tenant_id',
                                             'ctxt.project_id']),
        user_id=dict(type='text', fields=['payload.user_id',
                                          'ctxt.user_id']),
        # TODO(dikonoor):tenant_id is old terminology and should
        # be deprecated
        tenant_id=dict(type='text', fields=['payload.tenant_id',
                                            'ctxt.project_id']),
    )

    def __init__(self, definition_cfg, trait_plugin_mgr, raw_levels):
        self._included_types = []
        self._excluded_types = []
        self.traits = dict()
        self.cfg = definition_cfg
        self.raw_levels = raw_levels

        try:
            event_type = definition_cfg['event_type']
            traits = definition_cfg['traits']
        except KeyError as err:
            raise declarative.EventDefinitionException(
                _("Required field %s not specified") % err.args[0], self.cfg)

        if isinstance(event_type, str):
            event_type = [event_type]

        for t in event_type:
            if t.startswith('!'):
                self._excluded_types.append(t[1:])
            else:
                self._included_types.append(t)

        if self._excluded_types and not self._included_types:
            self._included_types.append('*')

        for trait_name in self.DEFAULT_TRAITS:
            self.traits[trait_name] = TraitDefinition(
                trait_name,
                self.DEFAULT_TRAITS[trait_name],
                trait_plugin_mgr)
        for trait_name in traits:
            self.traits[trait_name] = TraitDefinition(
                trait_name,
                traits[trait_name],
                trait_plugin_mgr)

    def included_type(self, event_type):
        for t in self._included_types:
            if fnmatch.fnmatch(event_type, t):
                return True
        return False

    def excluded_type(self, event_type):
        for t in self._excluded_types:
            if fnmatch.fnmatch(event_type, t):
                return True
        return False

    def match_type(self, event_type):
        return (self.included_type(event_type)
                and not self.excluded_type(event_type))

    @property
    def is_catchall(self):
        return '*' in self._included_types and not self._excluded_types

    def to_event(self, priority, notification_body):
        event_type = notification_body['event_type']
        message_id = notification_body['metadata']['message_id']
        when = timeutils.normalize_time(timeutils.parse_isotime(
            notification_body['metadata']['timestamp']))

        traits = (self.traits[t].to_trait(notification_body)
                  for t in self.traits)
        # Only accept non-None value traits ...
        traits = [trait for trait in traits if trait is not None]
        raw = notification_body if priority in self.raw_levels else {}
        event = models.Event(message_id, event_type, when, traits, raw)
        return event


class NotificationEventsConverter:
    """Notification Event Converter

    The NotificationEventsConverter handles the conversion of Notifications
    from openstack systems into Ceilometer Events.

    The conversion is handled according to event definitions in a config file.

    The config is a list of event definitions. Order is significant, a
    notification will be processed according to the LAST definition that
    matches it's event_type. (We use the last matching definition because that
    allows you to use YAML merge syntax in the definitions file.)
    Each definition is a dictionary with the following keys (all are
    required):

    - event_type: this is a list of notification event_types this definition
      will handle. These can be wildcarded with unix shell glob (not regex!)
      wildcards.
      An exclusion listing (starting with a '!') will exclude any types listed
      from matching. If ONLY exclusions are listed, the definition will match
      anything not matching the exclusions.
      This item can also be a string, which will be taken as equivalent to 1
      item list.

                Examples:

                * ['compute.instance.exists'] will only match
                  compute.instance.exists notifications
                * "compute.instance.exists"   Same as above.
                * ["image.create", "image.delete"]  will match
                  image.create and image.delete, but not anything else.
                * "compute.instance.*" will match
                  compute.instance.create.start but not image.upload
                * ['*.start','*.end', '!scheduler.*'] will match
                  compute.instance.create.start, and image.delete.end,
                  but NOT compute.instance.exists or
                  scheduler.run_instance.start
                * '!image.*' matches any notification except image
                  notifications.
                * ['*', '!image.*']  same as above.

    - traits: (dict) The keys are trait names, the values are the trait
      definitions. Each trait definition is a dictionary with the following
      keys:

      - type (optional): The data type for this trait. (as a string)
        Valid options are: 'text', 'int', 'float' and 'datetime', defaults to
        'text' if not specified.
      - fields:  a path specification for the field(s) in the notification you
        wish to extract. The paths can be specified with a dot syntax
        (e.g. 'payload.host') or dictionary syntax (e.g. 'payload[host]') is
        also supported.
        In either case, if the key for the field you are looking for contains
        special characters, like '.', it will need to be quoted (with double
        or single quotes) like so::

         "payload.image_meta.'org.openstack__1__architecture'"

        The syntax used for the field specification is a variant of JSONPath,
        and is fairly flexible.
        (see: https://github.com/kennknowles/python-jsonpath-rw for more info)
        Specifications can be written to match multiple possible fields, the
        value for the trait will be derived from the matching fields that
        exist and have a non-null (i.e. is not None) values in the
        notification.
        By default the value will be the first such field. (plugins can alter
        that, if they wish)

        This configuration value is normally a string, for convenience, it can
        be specified as a list of specifications, which will be OR'ed together
        (a union query in jsonpath terms)
    - plugin (optional): (dictionary) with the following keys:

      - name: (string) name of a plugin to load
      - parameters: (optional) Dictionary of keyword args to pass
        to the plugin on initialization. See documentation on each plugin to
        see what arguments it accepts.

      For convenience, this value can also be specified as a string, which is
      interpreted as a plugin name, which will be loaded with no parameters.

    """

    def __init__(self, conf, events_config, trait_plugin_mgr):
        self.conf = conf

        raw_levels = [level.lower() for level in self.conf.event.store_raw]
        self.definitions = [
            EventDefinition(event_def, trait_plugin_mgr, raw_levels)
            for event_def in reversed(events_config)]

        add_catchall = not self.conf.event.drop_unmatched_notifications
        if add_catchall and not any(d.is_catchall for d in self.definitions):
            event_def = dict(event_type='*', traits={})
            self.definitions.append(EventDefinition(event_def,
                                                    trait_plugin_mgr,
                                                    raw_levels))

    def to_event(self, priority, notification_body):
        event_type = notification_body['event_type']
        message_id = notification_body['metadata']['message_id']
        edef = None
        for d in self.definitions:
            if d.match_type(event_type):
                edef = d
                break

        if edef is None:
            if self.conf.event.drop_unmatched_notifications:
                msg_level = log.DEBUG
            else:
                # If drop_unmatched_notifications is False, this should
                # never happen. (mdragon)
                msg_level = log.ERROR
            LOG.log(msg_level,
                    'Dropping Notification %(type)s (uuid:%(msgid)s)',
                    dict(type=event_type, msgid=message_id))
            return None

        return edef.to_event(priority, notification_body)


def setup_events(conf, trait_plugin_mgr):
    """Setup the event definitions from yaml config file."""

    return NotificationEventsConverter(
        conf,
        declarative.load_definitions(
            conf, [], conf.event.definitions_cfg_file,
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                '..', 'pipeline', 'data', 'event_definitions.yaml')),
        trait_plugin_mgr)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log
from stevedore import extension

from ceilometer import agent
from ceilometer.pipeline import base

LOG = log.getLogger(__name__)


class SampleEndpoint(base.NotificationEndpoint):

    def info(self, notifications):
        """Convert message at info level to Ceilometer sample.

        :param notifications: list of notifications
        """
        return self.process_notifications('info', notifications)

    def sample(self, notifications):
        """Convert message at sample level to Ceilometer Event.

        :param notifications: list of notifications
        """
        return self.process_notifications('sample', notifications)

    def process_notifications(self, priority, notifications):
        for message in notifications:
            try:
                LOG.debug("Processing sample notification [%s] for publisher "
                          "[%s] with priority [%s] using the agent [%s].",
                          message, self.publisher, priority, self)
                with self.publisher as p:
                    p(list(self.build_sample(message)))
            except Exception:
                LOG.error('Fail to process notification message [%s]',
                          message, exc_info=True)
                raise

    def build_sample(notification):
        """Build sample from provided notification."""
        pass


class SampleSource(base.PipelineSource):
    """Represents a source of samples.

    In effect it is a set of notification handlers processing
    samples for a set of matching meters. Each source encapsulates meter name
    matching and mapping to one or more sinks for publication.
    """

    def __init__(self, cfg):
        super().__init__(cfg)
        try:
            self.meters = cfg['meters']
        except KeyError:
            raise base.PipelineException("Missing meters value", cfg)
        try:
            self.check_source_filtering(self.meters, 'meters')
        except agent.SourceException as err:
            raise base.PipelineException(err.msg, cfg)

    def support_meter(self, meter_name):
        return self.is_supported(self.meters, meter_name)


class SampleSink(base.Sink):

    def publish_samples(self, samples):
        """Push samples into pipeline for publishing.

        :param samples: Sample list.
        """

        if samples:
            for p in self.publishers:
                try:
                    p.publish_samples(samples)
                except Exception:
                    LOG.error("Pipeline %(pipeline)s: Continue after "
                              "error from publisher %(pub)s",
                              {'pipeline': self, 'pub': p},
                              exc_info=True)

    @staticmethod
    def flush():
        pass


class SamplePipeline(base.Pipeline):
    """Represents a pipeline for Samples."""

    def _validate_volume(self, s):
        volume = s.volume
        if volume is None:
            LOG.warning(
                'metering data %(counter_name)s for %(resource_id)s '
                '@ %(timestamp)s has no volume (volume: None), the sample will'
                ' be dropped',
                {'counter_name': s.name,
                 'resource_id': s.resource_id,
                 'timestamp': s.timestamp if s.timestamp else 'NO TIMESTAMP'}
            )
            return False
        if not isinstance(volume, (int, float)):
            try:
                volume = float(volume)
            except ValueError:
                LOG.warning(
                    'metering data %(counter_name)s for %(resource_id)s '
                    '@ %(timestamp)s has volume which is not a number '
                    '(volume: %(counter_volume)s), the sample will be dropped',
                    {'counter_name': s.name,
                     'resource_id': s.resource_id,
                     'timestamp': (
                         s.timestamp if s.timestamp else 'NO TIMESTAMP'),
                     'counter_volume': volume}
                )
                return False
        return True

    def publish_data(self, samples):
        if not isinstance(samples, list):
            samples = [samples]
        supported = [s for s in samples if self.supported(s)
                     and self._validate_volume(s)]
        self.sink.publish_samples(supported)

    def supported(self, sample):
        return self.source.support_meter(sample.name)


class SamplePipelineManager(base.PipelineManager):

    pm_type = 'sample'
    pm_pipeline = SamplePipeline
    pm_source = SampleSource
    pm_sink = SampleSink

    def __init__(self, conf):
        super().__init__(
            conf, conf.pipeline_cfg_file)

    def get_main_endpoints(self):
        exts = extension.ExtensionManager(
            namespace='ceilometer.sample.endpoint',
            invoke_on_load=True,
            invoke_args=(self.conf, self.publisher()))
        return [ext.obj for ext in exts]



#
# Copyright 2016 IBM
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ceilometer.publisher import http
from ceilometer import sample


class PrometheusPublisher(http.HttpPublisher):
    """Publish metering data to Prometheus Pushgateway endpoint

    This dispatcher inherits from all options of the http dispatcher.

    To use this publisher for samples, add the following section to the
    /etc/ceilometer/pipeline.yaml file or simply add it to an existing
    pipeline::

          - name: meter_file
            meters:
                - "*"
            publishers:
                - prometheus://mypushgateway/metrics/job/ceilometer

    """

    HEADERS = {'Content-type': 'plain/text'}

    def publish_samples(self, samples):
        """Send a metering message for publishing

        :param samples: Samples from pipeline after transformation
        """
        if not samples:
            return

        data = ""
        doc_done = set()
        for s in samples:
            # NOTE(sileht): delta can't be converted into prometheus data
            # format so don't set the metric type for it
            metric_type = None
            if s.type == sample.TYPE_CUMULATIVE:
                metric_type = "counter"
            elif s.type == sample.TYPE_GAUGE:
                metric_type = "gauge"

            curated_sname = s.name.replace(".", "_")

            if metric_type and curated_sname not in doc_done:
                data += f"# TYPE {curated_sname} {metric_type}\n"
                doc_done.add(curated_sname)

            # NOTE(sileht): prometheus pushgateway doesn't allow to push
            # timestamp_ms
            #
            # timestamp_ms = (
            #     s.get_iso_timestamp().replace(tzinfo=None) -
            #     datetime.utcfromtimestamp(0)
            # ).total_seconds() * 1000
            # data += '%s{resource_id="%s"} %s %d\n' % (
            #     curated_sname, s.resource_id, s.volume, timestamp_ms)

            data += '%s{resource_id="%s", user_id="%s", project_id="%s"}' \
                ' %s\n' % (curated_sname, s.resource_id, s.user_id,
                           s.project_id, s.volume)

        self._do_post(data)

    @staticmethod
    def publish_events(events):
        raise NotImplementedError



#
# Copyright 2014-2015 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from collections import defaultdict
import fnmatch
import hashlib
import itertools
import json
import operator
import os
import threading

from gnocchiclient import exceptions as gnocchi_exc
from keystoneauth1 import exceptions as ka_exceptions
from oslo_log import log
from oslo_utils import strutils
from oslo_utils import timeutils
from stevedore import extension
import tenacity
from urllib import parse as urlparse

from ceilometer import cache_utils
from ceilometer import declarative
from ceilometer import gnocchi_client
from ceilometer.i18n import _
from ceilometer import keystone_client
from ceilometer import publisher

LOG = log.getLogger(__name__)


EVENT_CREATE, EVENT_UPDATE, EVENT_DELETE = ("create", "update", "delete")


class ResourcesDefinition:

    MANDATORY_FIELDS = {'resource_type': str,
                        'metrics': (dict, list)}

    MANDATORY_EVENT_FIELDS = {'id': str}

    def __init__(self, definition_cfg, archive_policy_default,
                 archive_policy_override, plugin_manager):
        self.cfg = definition_cfg

        self._check_required_and_types(self.MANDATORY_FIELDS, self.cfg)

        if self.support_events():
            self._check_required_and_types(self.MANDATORY_EVENT_FIELDS,
                                           self.cfg['event_attributes'])

        self._attributes = {}
        for name, attr_cfg in self.cfg.get('attributes', {}).items():
            self._attributes[name] = declarative.Definition(name, attr_cfg,
                                                            plugin_manager)

        self._event_attributes = {}
        for name, attr_cfg in self.cfg.get('event_attributes', {}).items():
            self._event_attributes[name] = declarative.Definition(
                name, attr_cfg, plugin_manager)

        self.metrics = {}

        # NOTE(sileht): Convert old list to new dict format
        if isinstance(self.cfg['metrics'], list):
            values = [None] * len(self.cfg['metrics'])
            self.cfg['metrics'] = dict(zip(self.cfg['metrics'], values))

        for m, extra in self.cfg['metrics'].items():
            if not extra:
                extra = {}

            if not extra.get("archive_policy_name"):
                extra["archive_policy_name"] = archive_policy_default

            if archive_policy_override:
                extra["archive_policy_name"] = archive_policy_override

            # NOTE(sileht): For backward compat, this is after the override to
            # preserve the weird previous behavior. We don't really care as we
            # deprecate it.
            if 'archive_policy' in self.cfg:
                LOG.warning("archive_policy '%s' for a resource-type (%s) is "
                            "deprecated, set it for each metric instead.",
                            self.cfg["archive_policy"],
                            self.cfg["resource_type"])
                extra["archive_policy_name"] = self.cfg['archive_policy']

            self.metrics[m] = extra

    @staticmethod
    def _check_required_and_types(expected, definition):
        for field, field_types in expected.items():
            if field not in definition:
                raise declarative.ResourceDefinitionException(
                    _("Required field %s not specified") % field, definition)
            if not isinstance(definition[field], field_types):
                raise declarative.ResourceDefinitionException(
                    _("Required field %(field)s should be a %(type)s") %
                    {'field': field, 'type': field_types}, definition)

    @staticmethod
    def _ensure_list(value):
        if isinstance(value, list):
            return value
        return [value]

    def support_events(self):
        for e in ["event_create", "event_delete", "event_update"]:
            if e in self.cfg:
                return True
        return False

    def event_match(self, event_type):
        for e in self._ensure_list(self.cfg.get('event_create', [])):
            if fnmatch.fnmatch(event_type, e):
                return EVENT_CREATE
        for e in self._ensure_list(self.cfg.get('event_delete', [])):
            if fnmatch.fnmatch(event_type, e):
                return EVENT_DELETE
        for e in self._ensure_list(self.cfg.get('event_update', [])):
            if fnmatch.fnmatch(event_type, e):
                return EVENT_UPDATE

    def sample_attributes(self, sample):
        attrs = {}
        sample_dict = sample.as_dict()
        for name, definition in self._attributes.items():
            value = definition.parse(sample_dict)
            if value is not None:
                attrs[name] = value
        return attrs

    def event_attributes(self, event):
        attrs = {'type': self.cfg['resource_type']}
        traits = {trait.name: trait.value for trait in event.traits}
        for attr, field in self.cfg.get('event_attributes', {}).items():
            value = traits.get(field)
            if value is not None:
                attrs[attr] = value
        return attrs


class LockedDefaultDict(defaultdict):
    """defaultdict with lock to handle threading

    Dictionary only deletes if nothing is accessing dict and nothing is holding
    lock to be deleted. If both cases are not true, it will skip delete.
    """

    def __init__(self, *args, **kwargs):
        self.lock = threading.Lock()
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        with self.lock:
            return super().__getitem__(key)

    def pop(self, key, *args):
        with self.lock:
            key_lock = super().__getitem__(key)
            if key_lock.acquire(False):
                try:
                    super().pop(key, *args)
                finally:
                    key_lock.release()


class GnocchiPublisher(publisher.ConfigPublisherBase):
    """Publisher class for recording metering data into the Gnocchi service.

    The publisher class records each meter into the gnocchi service
    configured in Ceilometer pipeline file. An example target may
    look like the following:

      gnocchi://?archive_policy=low&filter_project=gnocchi

    To disable filtering the Gnocchi project, use the following option:

      gnocchi://?enable_filter_project=false
    """

    def __init__(self, conf, parsed_url):
        super().__init__(conf, parsed_url)
        # TODO(jd) allow to override Gnocchi endpoint via the host in the URL
        options = urlparse.parse_qs(parsed_url.query)

        self.enable_filter_project = strutils.bool_from_string(
            options.get('enable_filter_project', ['true'])[-1])
        if self.enable_filter_project:
            self.filter_project = options.get('filter_project',
                                              ['service'])[-1]
            self.filter_domain = options.get('filter_domain', ['Default'])[-1]
        else:
            LOG.debug(
                "Filtering Gnocchi project is disabled, "
                "ignoring filter_project option")
            self.filter_project = None
            self.filter_domain = None

        resources_definition_file = options.get(
            'resources_definition_file', ['gnocchi_resources.yaml'])[-1]

        archive_policy_override = options.get('archive_policy', [None])[-1]
        self.resources_definition, self.archive_policies_definition = (
            self._load_definitions(conf, archive_policy_override,
                                   resources_definition_file))
        self.metric_map = {metric: rd for rd in self.resources_definition
                           for metric in rd.metrics}

        timeout = options.get('timeout', [6.05])[-1]
        self._ks_client = keystone_client.get_client(conf)

        # NOTE(cdent): The default cache backend is a real but
        # noop backend. We don't want to use that here because
        # we want to avoid the cache pathways entirely if the
        # cache has not been configured explicitly.
        self.cache = cache_utils.get_client(conf)

        self._gnocchi_project_id = None
        self._gnocchi_project_id_lock = threading.Lock()
        self._gnocchi_resource_lock = LockedDefaultDict(threading.Lock)

        try:
            self._gnocchi = self._get_gnocchi_client(conf, timeout)
        except tenacity.RetryError as e:
            raise e.last_attempt._exception from None

        self._already_logged_event_types = set()
        self._already_logged_metric_names = set()

        self._already_configured_archive_policies = False

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(10),
        wait=tenacity.wait_fixed(5),
        retry=(
            tenacity.retry_if_exception_type(ka_exceptions.ServiceUnavailable)
            | tenacity.retry_if_exception_type(ka_exceptions.DiscoveryFailure)
            | tenacity.retry_if_exception_type(ka_exceptions.ConnectTimeout)
        ),
        reraise=False)
    def _get_gnocchi_client(self, conf, timeout):
        return gnocchi_client.get_gnocchiclient(conf, request_timeout=timeout)

    @staticmethod
    def _load_definitions(conf, archive_policy_override,
                          resources_definition_file):
        plugin_manager = extension.ExtensionManager(
            namespace='ceilometer.event.trait_plugin')
        data = declarative.load_definitions(
            conf, {}, resources_definition_file,
            os.path.join(os.path.dirname(os.path.abspath(__file__)),
                         'data', 'gnocchi_resources.yaml'))

        archive_policy_default = data.get("archive_policy_default",
                                          "ceilometer-low")
        resource_defs = []
        for resource in data.get('resources', []):
            try:
                resource_defs.append(ResourcesDefinition(
                    resource,
                    archive_policy_default,
                    archive_policy_override,
                    plugin_manager))
            except Exception:
                LOG.error("Failed to load resource due to error",
                          exc_info=True)
        return resource_defs, data.get("archive_policies", [])

    def ensures_archives_policies(self):
        if not self._already_configured_archive_policies:
            for ap in self.archive_policies_definition:
                try:
                    self._gnocchi.archive_policy.create(ap)
                except gnocchi_exc.ArchivePolicyAlreadyExists:
                    # created in the meantime by another worker
                    pass
            self._already_configured_archive_policies = True

    @property
    def gnocchi_project_id(self):
        if not self.enable_filter_project:
            return None
        if self._gnocchi_project_id is not None:
            return self._gnocchi_project_id
        with self._gnocchi_project_id_lock:
            if self._gnocchi_project_id is None:
                if not self.filter_project:
                    LOG.debug(
                        "Multiple executions were locked on "
                        "self._gnocchi_project_id_lock`. This execution "
                        "should no call `_internal_gnocchi_project_discovery` "
                        "as `self.filter_project` is None.")
                    return None
                try:
                    domain = self._ks_client.domains.find(
                        name=self.filter_domain)
                    project = self._ks_client.projects.find(
                        name=self.filter_project,
                        domain_id=domain.id)
                except ka_exceptions.NotFound:
                    LOG.warning('Filtered project [%s] not found in keystone, '
                                'ignoring the filter_project option',
                                self.filter_project)

                    self.filter_project = None
                    return None
                except Exception:
                    LOG.exception('Failed to retrieve filtered project [%s].',
                                  self.filter_project)
                    raise
                self._gnocchi_project_id = project.id
                LOG.debug("Filtered project [%s] found with ID [%s].",
                          self.filter_project, self._gnocchi_project_id)
            return self._gnocchi_project_id

    def _is_swift_account_sample(self, sample):
        try:
            return (self.metric_map[sample.name].cfg['resource_type']
                    == 'swift_account')
        except KeyError:
            return False

    def _is_gnocchi_activity(self, sample):
        return (self.filter_project and self.gnocchi_project_id and (
            # avoid anything from the user used by gnocchi
            sample.project_id == self.gnocchi_project_id or
            # avoid anything in the swift account used by gnocchi
            (sample.resource_id == self.gnocchi_project_id and
             self._is_swift_account_sample(sample))
        ))

    def _get_resource_definition_from_event(self, event_type):
        for rd in self.resources_definition:
            operation = rd.event_match(event_type)
            if operation:
                return rd, operation

    def filter_gnocchi_activity_openstack(self, samples):
        """Skip sample generated by gnocchi itself

        This method will filter out the samples that are generated by
        Gnocchi itself.
        """
        if not self.enable_filter_project:
            LOG.debug("Filtering Gnocchi activities is disabled, "
                      "samples [%s] will not be filtered.", samples)
            return samples
        filtered_samples = []
        for sample in samples:
            if not self._is_gnocchi_activity(sample):
                filtered_samples.append(sample)
                LOG.debug("Sample [%s] is not a Gnocchi activity; therefore, "
                          "we do not filter it out and push it to Gnocchi.",
                          sample)
            else:
                LOG.debug("Sample [%s] is a Gnocchi activity; therefore, "
                          "we filter it out and do not push it to Gnocchi.",
                          sample)
        return filtered_samples

    def publish_samples(self, data):
        self.ensures_archives_policies()

        data = self.filter_gnocchi_activity_openstack(data)

        def value_to_sort(object_to_sort):
            value = object_to_sort.resource_id
            if not value:
                LOG.debug("Resource ID was not defined for sample data [%s]. "
                          "Therefore, we will use an empty string as the "
                          "resource ID.", object_to_sort)
                value = ''

            return value

        data.sort(key=value_to_sort)
        resource_grouped_samples = itertools.groupby(
            data, key=operator.attrgetter('resource_id'))

        gnocchi_data = {}
        measures = {}
        for resource_id, samples_of_resource in resource_grouped_samples:
            for sample in samples_of_resource:
                metric_name = sample.name
                LOG.debug("Processing sample [%s] for resource ID [%s].",
                          sample, resource_id)
                rd = self.metric_map.get(metric_name)
                if rd is None:
                    if metric_name not in self._already_logged_metric_names:
                        LOG.warning("metric %s is not handled by Gnocchi",
                                    metric_name)
                        self._already_logged_metric_names.add(metric_name)
                    continue

                # NOTE(sileht): / is forbidden by Gnocchi
                resource_id = resource_id.replace('/', '_')
                if resource_id not in gnocchi_data:
                    gnocchi_data[resource_id] = {
                        'resource_type': rd.cfg['resource_type'],
                        'resource': {"id": resource_id,
                                     "user_id": sample.user_id,
                                     "project_id": sample.project_id}}

                gnocchi_data[resource_id].setdefault(
                    "resource_extra", {}).update(rd.sample_attributes(sample))
                measures.setdefault(resource_id, {}).setdefault(
                    metric_name,
                    {"measures": [],
                     "archive_policy_name":
                     rd.metrics[metric_name]["archive_policy_name"],
                     "unit": sample.unit}
                )["measures"].append(
                    {'timestamp': sample.timestamp,
                     'value': sample.volume}
                )

        try:
            self.batch_measures(measures, gnocchi_data)
        except gnocchi_exc.ClientException as e:
            LOG.error("Gnocchi client exception while pushing measures [%s] "
                      "for gnocchi data [%s]: [%s].", measures, gnocchi_data,
                      str(e))
        except Exception as e:
            LOG.error("Unexpected exception while pushing measures [%s] for "
                      "gnocchi data [%s]: [%s].", measures, gnocchi_data,
                      str(e), exc_info=True)

        for info in gnocchi_data.values():
            resource = info["resource"]
            resource_type = info["resource_type"]
            resource_extra = info["resource_extra"]
            if not resource_extra:
                continue
            try:
                self._if_not_cached(resource_type, resource['id'],
                                    resource_extra)
            except gnocchi_exc.ClientException as e:
                LOG.error("Gnocchi client exception updating resource type "
                          "[%s] with ID [%s] for resource data [%s]: [%s].",
                          resource_type, resource.get('id'), resource_extra,
                          str(e))
            except Exception as e:
                LOG.error("Unexpected exception updating resource type [%s] "
                          "with ID [%s] for resource data [%s]: [%s].",
                          resource_type, resource.get('id'), resource_extra,
                          str(e), exc_info=True)

    @staticmethod
    def _extract_resources_from_error(e, resource_infos):
        resource_ids = {r['original_resource_id']
                        for r in e.message['detail']}
        return [(resource_infos[rid]['resource_type'],
                 resource_infos[rid]['resource'],
                 resource_infos[rid]['resource_extra'])
                for rid in resource_ids]

    def batch_measures(self, measures, resource_infos):
        # NOTE(sileht): We don't care about error here, we want
        # resources metadata always been updated
        try:
            LOG.debug("Executing batch resource metrics measures for resource "
                      "[%s] and measures [%s].", resource_infos, measures)

            self._gnocchi.metric.batch_resources_metrics_measures(
                measures, create_metrics=True)
        except gnocchi_exc.BadRequest as e:
            if not isinstance(e.message, dict):
                raise
            if e.message.get('cause') != 'Unknown resources':
                raise

            resources = self._extract_resources_from_error(e, resource_infos)
            for resource_type, resource, resource_extra in resources:
                try:
                    resource.update(resource_extra)
                    self._create_resource(resource_type, resource)
                except gnocchi_exc.ResourceAlreadyExists:
                    # NOTE(sileht): resource created in the meantime
                    pass
                except gnocchi_exc.ClientException as e:
                    LOG.error('Error creating resource %(id)s: %(err)s',
                              {'id': resource['id'], 'err': str(e)})
                    # We cannot post measures for this resource
                    # and we can't patch it later
                    del measures[resource['id']]
                    del resource_infos[resource['id']]
                else:
                    if self.cache and resource_extra:
                        self.cache.set(resource['id'],
                                       self._hash_resource(resource_extra))

            # NOTE(sileht): we have created missing resources/metrics,
            # now retry to post measures
            self._gnocchi.metric.batch_resources_metrics_measures(
                measures, create_metrics=True)

        LOG.debug(
            "%d measures posted against %d metrics through %d resources",
            sum(len(m["measures"])
                for rid in measures
                for m in measures[rid].values()),
            sum(len(m) for m in measures.values()), len(resource_infos))

    def _create_resource(self, resource_type, resource):
        self._gnocchi.resource.create(resource_type, resource)
        LOG.debug('Resource %s created', resource["id"])

    def _update_resource(self, resource_type, res_id, resource_extra):
        self._gnocchi.resource.update(resource_type, res_id, resource_extra)
        LOG.debug('Resource %s updated', res_id)

    def _if_not_cached(self, resource_type, res_id, resource_extra):
        if self.cache:
            attribute_hash = self._hash_resource(resource_extra)
            if self._resource_cache_diff(res_id, attribute_hash):
                with self._gnocchi_resource_lock[res_id]:
                    # NOTE(luogangyi): there is a possibility that the
                    # resource was already built in cache by another
                    # ceilometer-notification-agent when we get the lock here.
                    if self._resource_cache_diff(res_id, attribute_hash):
                        self._update_resource(resource_type, res_id,
                                              resource_extra)
                        self.cache.set(res_id, attribute_hash)
                    else:
                        LOG.debug('Resource cache hit for %s', res_id)
                self._gnocchi_resource_lock.pop(res_id, None)
            else:
                LOG.debug('Resource cache hit for %s', res_id)
        else:
            self._update_resource(resource_type, res_id, resource_extra)

    @staticmethod
    def _hash_resource(resource):
        data = {k: v for k, v in resource.items() if k != 'metrics'}
        payload = json.dumps(data, sort_keys=True, separators=(',', ':'))
        return hashlib.blake2b(
            payload.encode(), usedforsecurity=False
        ).hexdigest()

    def _resource_cache_diff(self, key, attribute_hash):
        cached_hash = self.cache.get(key)
        return not cached_hash or cached_hash != attribute_hash

    def publish_events(self, events):
        for event in events:
            rd = self._get_resource_definition_from_event(event.event_type)
            if not rd:
                if event.event_type not in self._already_logged_event_types:
                    LOG.debug("No gnocchi definition for event type: %s",
                              event.event_type)
                    self._already_logged_event_types.add(event.event_type)
                continue

            rd, operation = rd
            if operation == EVENT_DELETE:
                self._delete_event(rd, event)
            if operation == EVENT_CREATE:
                self._create_event(rd, event)
            if operation == EVENT_UPDATE:
                self._update_event(rd, event)

    def _update_event(self, rd, event):
        resource = rd.event_attributes(event)
        associated_resources = rd.cfg.get('event_associated_resources', {})

        if associated_resources:
            to_update = itertools.chain([resource], *[
                self._search_resource(resource_type, query % resource['id'])
                for resource_type, query in associated_resources.items()
            ])
        else:
            to_update = [resource]

        for resource in to_update:
            self._set_update_attributes(resource)

    def _delete_event(self, rd, event):
        ended_at = timeutils.utcnow().isoformat()

        resource = rd.event_attributes(event)
        associated_resources = rd.cfg.get('event_associated_resources', {})

        if associated_resources:
            to_end = itertools.chain([resource], *[
                self._search_resource(resource_type, query % resource['id'])
                for resource_type, query in associated_resources.items()
            ])
        else:
            to_end = [resource]

        for resource in to_end:
            self._set_ended_at(resource, ended_at)

    def _create_event(self, rd, event):
        resource = rd.event_attributes(event)
        resource_type = resource.pop('type')

        try:
            self._create_resource(resource_type, resource)
        except gnocchi_exc.ResourceAlreadyExists:
            LOG.debug("Create event received on existing resource (%s), "
                      "ignore it.", resource['id'])
        except Exception:
            LOG.error("Failed to create resource %s", resource,
                      exc_info=True)

    def _search_resource(self, resource_type, query):
        try:
            return self._gnocchi.resource.search(
                resource_type, json.loads(query))
        except Exception:
            LOG.error("Fail to search resource type %(resource_type)s "
                      "with '%(query)s'",
                      {'resource_type': resource_type, 'query': query},
                      exc_info=True)
        return []

    def _set_update_attributes(self, resource):
        resource_id = resource.pop('id')
        resource_type = resource.pop('type')

        try:
            self._if_not_cached(resource_type, resource_id, resource)
        except gnocchi_exc.ResourceNotFound:
            LOG.debug("Update event received on unexisting resource (%s), "
                      "ignore it.", resource_id)
        except Exception:
            LOG.error("Fail to update the resource %s", resource,
                      exc_info=True)

    def _set_ended_at(self, resource, ended_at):
        try:
            self._gnocchi.resource.update(resource['type'], resource['id'],
                                          {'ended_at': ended_at})
        except gnocchi_exc.ResourceNotFound:
            LOG.debug("Delete event received on unexisting resource (%s), "
                      "ignore it.", resource['id'])
        except Exception:
            LOG.error("Fail to update the resource %s", resource,
                      exc_info=True)
        LOG.debug('Resource %(resource_id)s ended at %(ended_at)s',
                  {'resource_id': resource["id"], 'ended_at': ended_at})



#
# Copyright 2016 IBM
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Tests for ceilometer/publisher/http.py"""

import datetime
from unittest import mock
import uuid

from oslo_utils import timeutils
from oslotest import base
import requests
from urllib import parse as urlparse

from ceilometer.event import models as event
from ceilometer.publisher import http
from ceilometer import sample
from ceilometer import service


class TestHttpPublisher(base.BaseTestCase):

    resource_id = str(uuid.uuid4())

    sample_data = [
        sample.Sample(
            name='alpha',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id=resource_id,
            timestamp=timeutils.utcnow().isoformat(),
            resource_metadata={'name': 'TestPublish'},
        ),
        sample.Sample(
            name='beta',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id=resource_id,
            timestamp=timeutils.utcnow().isoformat(),
            resource_metadata={'name': 'TestPublish'},
        ),
        sample.Sample(
            name='gamma',
            type=sample.TYPE_CUMULATIVE,
            unit='',
            volume=1,
            user_id='test',
            project_id='test',
            resource_id=resource_id,
            timestamp=datetime.datetime.now().isoformat(),
            resource_metadata={'name': 'TestPublish'},
        ),
    ]

    event_data = [event.Event(
        message_id=str(uuid.uuid4()), event_type='event_%d' % i,
        generated=timeutils.utcnow().isoformat(),
        traits=[], raw={'payload': {'some': 'aa'}}) for i in range(3)]

    def setUp(self):
        super().setUp()
        self.CONF = service.prepare_service([], [])

    def test_http_publisher_config(self):
        """Test publisher config parameters."""
        # invalid hostname, the given url, results in an empty hostname
        parsed_url = urlparse.urlparse('http:/aaa.bb/path')
        self.assertRaises(ValueError, http.HttpPublisher,
                          self.CONF, parsed_url)

        # invalid port
        parsed_url = urlparse.urlparse('http://aaa:bb/path')
        self.assertRaises(ValueError, http.HttpPublisher,
                          self.CONF, parsed_url)

        parsed_url = urlparse.urlparse('http://localhost:90/path1')
        publisher = http.HttpPublisher(self.CONF, parsed_url)
        # By default, timeout and retry_count should be set to 5 and 2
        # respectively
        self.assertEqual(5, publisher.timeout)
        self.assertEqual(2, publisher.max_retries)

        parsed_url = urlparse.urlparse('http://localhost:90/path1?'
                                       'timeout=19&max_retries=4')
        publisher = http.HttpPublisher(self.CONF, parsed_url)
        self.assertEqual(19, publisher.timeout)
        self.assertEqual(4, publisher.max_retries)

        parsed_url = urlparse.urlparse('http://localhost:90/path1?'
                                       'timeout=19')
        publisher = http.HttpPublisher(self.CONF, parsed_url)
        self.assertEqual(19, publisher.timeout)
        self.assertEqual(2, publisher.max_retries)

        parsed_url = urlparse.urlparse('http://localhost:90/path1?'
                                       'max_retries=6')
        publisher = http.HttpPublisher(self.CONF, parsed_url)
        self.assertEqual(5, publisher.timeout)
        self.assertEqual(6, publisher.max_retries)

    @mock.patch('ceilometer.publisher.http.LOG')
    def test_http_post_samples(self, thelog):
        """Test publisher post."""
        parsed_url = urlparse.urlparse('http://localhost:90/path1')
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        res = requests.Response()
        res.status_code = 200
        with mock.patch.object(requests.Session, 'post',
                               return_value=res) as m_req:
            publisher.publish_samples(self.sample_data)

        self.assertEqual(1, m_req.call_count)
        self.assertFalse(thelog.exception.called)

        res = requests.Response()
        res.status_code = 401
        with mock.patch.object(requests.Session, 'post',
                               return_value=res) as m_req:
            publisher.publish_samples(self.sample_data)

        self.assertEqual(1, m_req.call_count)
        self.assertTrue(thelog.exception.called)

    @mock.patch('ceilometer.publisher.http.LOG')
    def test_http_post_events(self, thelog):
        """Test publisher post."""
        parsed_url = urlparse.urlparse('http://localhost:90/path1')
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        res = requests.Response()
        res.status_code = 200
        with mock.patch.object(requests.Session, 'post',
                               return_value=res) as m_req:
            publisher.publish_events(self.event_data)

        self.assertEqual(1, m_req.call_count)
        self.assertFalse(thelog.exception.called)

        res = requests.Response()
        res.status_code = 401
        with mock.patch.object(requests.Session, 'post',
                               return_value=res) as m_req:
            publisher.publish_events(self.event_data)

        self.assertEqual(1, m_req.call_count)
        self.assertTrue(thelog.exception.called)

    @mock.patch('ceilometer.publisher.http.LOG')
    def test_http_post_empty_data(self, thelog):
        parsed_url = urlparse.urlparse('http://localhost:90/path1')
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        res = requests.Response()
        res.status_code = 200
        with mock.patch.object(requests.Session, 'post',
                               return_value=res) as m_req:
            publisher.publish_events([])

        self.assertEqual(0, m_req.call_count)
        self.assertTrue(thelog.debug.called)

    def _post_batch_control_test(self, method, data, batch):
        parsed_url = urlparse.urlparse('http://localhost:90/path1?'
                                       'batch=%s' % batch)
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        with mock.patch.object(requests.Session, 'post') as post:
            getattr(publisher, method)(data)
            self.assertEqual(1 if batch else 3, post.call_count)

    def test_post_batch_sample(self):
        self._post_batch_control_test('publish_samples', self.sample_data, 1)

    def test_post_no_batch_sample(self):
        self._post_batch_control_test('publish_samples', self.sample_data, 0)

    def test_post_batch_event(self):
        self._post_batch_control_test('publish_events', self.event_data, 1)

    def test_post_no_batch_event(self):
        self._post_batch_control_test('publish_events', self.event_data, 0)

    def test_post_verify_ssl_default(self):
        parsed_url = urlparse.urlparse('http://localhost:90/path1')
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        with mock.patch.object(requests.Session, 'post') as post:
            publisher.publish_samples(self.sample_data)
            self.assertTrue(post.call_args[1]['verify'])

    def test_post_verify_ssl_True(self):
        parsed_url = urlparse.urlparse('http://localhost:90/path1?'
                                       'verify_ssl=True')
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        with mock.patch.object(requests.Session, 'post') as post:
            publisher.publish_samples(self.sample_data)
            self.assertTrue(post.call_args[1]['verify'])

    def test_post_verify_ssl_False(self):
        parsed_url = urlparse.urlparse('http://localhost:90/path1?'
                                       'verify_ssl=False')
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        with mock.patch.object(requests.Session, 'post') as post:
            publisher.publish_samples(self.sample_data)
            self.assertFalse(post.call_args[1]['verify'])

    def test_post_verify_ssl_path(self):
        parsed_url = urlparse.urlparse('http://localhost:90/path1?'
                                       'verify_ssl=/path/to/cert.crt')
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        with mock.patch.object(requests.Session, 'post') as post:
            publisher.publish_samples(self.sample_data)
            self.assertEqual('/path/to/cert.crt', post.call_args[1]['verify'])

    def test_post_basic_auth(self):
        parsed_url = urlparse.urlparse(
            'http://alice:l00kingGla$$@localhost:90/path1?')
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        with mock.patch.object(requests.Session, 'post') as post:
            publisher.publish_samples(self.sample_data)
            self.assertEqual(('alice', 'l00kingGla$$'),
                             post.call_args[1]['auth'])

    def test_post_client_cert_auth(self):
        parsed_url = urlparse.urlparse('http://localhost:90/path1?'
                                       'clientcert=/path/to/cert.crt&'
                                       'clientkey=/path/to/cert.key')
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        with mock.patch.object(requests.Session, 'post') as post:
            publisher.publish_samples(self.sample_data)
            self.assertEqual(('/path/to/cert.crt', '/path/to/cert.key'),
                             post.call_args[1]['cert'])

    def test_post_raw_only(self):
        parsed_url = urlparse.urlparse('http://localhost:90/path1?raw_only=1')
        publisher = http.HttpPublisher(self.CONF, parsed_url)

        with mock.patch.object(requests.Session, 'post') as post:
            publisher.publish_events(self.event_data)
            self.assertEqual(
                '[{"some": "aa"}, {"some": "aa"}, {"some": "aa"}]',
                post.call_args[1]['data'])



#
# Copyright 2016 IBM
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from oslo_log import log
from oslo_utils import strutils
import requests
from requests import adapters
from urllib import parse as urlparse

from ceilometer import publisher

LOG = log.getLogger(__name__)


class HttpPublisher(publisher.ConfigPublisherBase):
    """Publish metering data to a http endpoint

    This publisher pushes metering data to a specified http endpoint. The
    endpoint should be configured in ceilometer pipeline configuration file.
    If the `timeout` and/or `max_retries` are not specified, the default
    `timeout` and `max_retries` will be set to 5 and 2 respectively. Additional
    parameters are:

        - ssl certificate verification can be disabled by setting `verify_ssl`
          to False
        - batching can be configured by `batch`
        - Basic authentication can be configured using the URL authentication
          scheme: http://username:password@example.com
        - For certificate authentication, `clientcert` and `clientkey` are the
          paths to the certificate and key files respectively. `clientkey` is
          only required if the clientcert file doesn't already contain the key.

    All of the parameters mentioned above get removed during processing,
    with the remaining portion of the URL being used as the actual endpoint.
    e.g. https://username:password@example.com/path?verify_ssl=False&q=foo
    will result in a call to https://example.com/path?q=foo

    To use this publisher for samples, add the following section to the
    /etc/ceilometer/pipeline.yaml file or simply add it to an existing
    pipeline::

          - name: meter_file
            meters:
                - "*"
            publishers:
                - http://host:80/path?timeout=1&max_retries=2&batch=False

    In the event_pipeline.yaml file, you can use the publisher in one of
    the sinks like the following:

          - name: event_sink
            publishers:
                - http://host:80/path?timeout=1&max_retries=2

    """

    HEADERS = {'Content-type': 'application/json'}

    def __init__(self, conf, parsed_url):
        super().__init__(conf, parsed_url)

        if not parsed_url.hostname:
            raise ValueError('The hostname of an endpoint for '
                             'HttpPublisher is required')

        # non-numeric port from the url string will cause a ValueError
        # exception when the port is read. Do a read to make sure the port
        # is valid, if not, ValueError will be thrown.
        parsed_url.port

        # Handling other configuration options in the query string
        params = urlparse.parse_qs(parsed_url.query)
        self.timeout = self._get_param(params, 'timeout', 5, int)
        self.max_retries = self._get_param(params, 'max_retries', 2, int)
        self.poster = (
            self._batch_post if strutils.bool_from_string(self._get_param(
                params, 'batch', True)) else self._individual_post)
        verify_ssl = self._get_param(params, 'verify_ssl', True)
        try:
            self.verify_ssl = strutils.bool_from_string(verify_ssl,
                                                        strict=True)
        except ValueError:
            self.verify_ssl = (verify_ssl or True)

        username = parsed_url.username
        password = parsed_url.password
        if username:
            self.client_auth = (username, password)
            netloc = parsed_url.netloc.replace(username + ':' + password + '@',
                                               '')
        else:
            self.client_auth = None
            netloc = parsed_url.netloc

        clientcert = self._get_param(params, 'clientcert', None)
        clientkey = self._get_param(params, 'clientkey', None)
        if clientcert:
            if clientkey:
                self.client_cert = (clientcert, clientkey)
            else:
                self.client_cert = clientcert
        else:
            self.client_cert = None

        self.raw_only = strutils.bool_from_string(
            self._get_param(params, 'raw_only', False))

        kwargs = {'max_retries': self.max_retries,
                  'pool_connections': conf.max_parallel_requests,
                  'pool_maxsize': conf.max_parallel_requests}
        self.session = requests.Session()

        if parsed_url.scheme in ["http", "https"]:
            scheme = parsed_url.scheme
        else:
            ssl = self._get_param(params, 'ssl', False)
            try:
                ssl = strutils.bool_from_string(ssl, strict=True)
            except ValueError:
                ssl = (ssl or False)
            scheme = "https" if ssl else "http"

        # authentication & config params have been removed, so use URL with
        # updated query string
        self.target = urlparse.urlunsplit([
            scheme,
            netloc,
            parsed_url.path,
            urlparse.urlencode(params, doseq=True),
            parsed_url.fragment])

        self.session.mount(self.target, adapters.HTTPAdapter(**kwargs))

        LOG.debug('HttpPublisher for endpoint %s is initialized!',
                  self.target)

    @staticmethod
    def _get_param(params, name, default_value, cast=None):
        try:
            return cast(params.pop(name)[-1]) if cast else params.pop(name)[-1]
        except (ValueError, TypeError, KeyError):
            LOG.debug('Default value %(value)s is used for %(name)s',
                      {'value': default_value, 'name': name})
            return default_value

    def _individual_post(self, data):
        for d in data:
            self._do_post(json.dumps(d))

    def _batch_post(self, data):
        if not data:
            LOG.debug('Data set is empty!')
            return
        self._do_post(json.dumps(data))

    def _do_post(self, data):
        LOG.trace('Message: %s', data)
        try:
            res = self.session.post(self.target, data=data,
                                    headers=self.HEADERS, timeout=self.timeout,
                                    auth=self.client_auth,
                                    cert=self.client_cert,
                                    verify=self.verify_ssl)
            res.raise_for_status()
            LOG.debug('Message posting to %s: status code %d.',
                      self.target, res.status_code)
        except requests.exceptions.HTTPError:
            LOG.exception('Status Code: %(code)s. '
                          'Failed to dispatch message: %(data)s',
                          {'code': res.status_code, 'data': data})

    def publish_samples(self, samples):
        """Send a metering message for publishing

        :param samples: Samples from pipeline after transformation
        """
        self.poster([sample.as_dict() for sample in samples])

    def publish_events(self, events):
        """Send an event message for publishing

        :param events: events from pipeline after transformation
        """
        if self.raw_only:
            data = [evt.as_dict()['raw']['payload'] for evt in events
                    if evt.as_dict().get('raw', {}).get('payload')]
        else:
            data = [event.serialize() for event in events]
        self.poster(data)



# Copyright 2012-2014 eNovance <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log
import oslo_messaging
from stevedore import extension

from ceilometer import agent
from ceilometer.event import converter
from ceilometer.pipeline import base

LOG = log.getLogger(__name__)


class EventEndpoint(base.NotificationEndpoint):

    event_types = []

    def __init__(self, conf, publisher):
        super().__init__(conf, publisher)
        LOG.debug('Loading event definitions')
        self.event_converter = converter.setup_events(
            conf,
            extension.ExtensionManager(
                namespace='ceilometer.event.trait_plugin'))

    def info(self, notifications):
        """Convert message at info level to Ceilometer Event.

        :param notifications: list of notifications
        """
        return self.process_notifications('info', notifications)

    def error(self, notifications):
        """Convert message at error level to Ceilometer Event.

        :param notifications: list of notifications
        """
        return self.process_notifications('error', notifications)

    def process_notifications(self, priority, notifications):
        for message in notifications:
            try:
                event = self.event_converter.to_event(priority, message)
                if event is not None:
                    with self.publisher as p:
                        p(event)
            except Exception:
                if not self.conf.notification.ack_on_event_error:
                    return oslo_messaging.NotificationResult.REQUEUE
                LOG.error('Fail to process a notification', exc_info=True)
        return oslo_messaging.NotificationResult.HANDLED


class EventSource(base.PipelineSource):
    """Represents a source of events.

    In effect it is a set of notification handlers capturing events for a set
    of matching notifications.
    """

    def __init__(self, cfg):
        super().__init__(cfg)
        self.events = cfg.get('events')
        try:
            self.check_source_filtering(self.events, 'events')
        except agent.SourceException as err:
            raise base.PipelineException(err.msg, cfg)

    def support_event(self, event_name):
        return self.is_supported(self.events, event_name)


class EventSink(base.Sink):

    def publish_events(self, events):
        if events:
            for p in self.publishers:
                try:
                    p.publish_events(events)
                except Exception:
                    LOG.error("Pipeline %(pipeline)s: %(status)s "
                              "after error from publisher %(pub)s",
                              {'pipeline': self,
                               'status': 'Continue' if
                               self.multi_publish else 'Exit', 'pub': p},
                              exc_info=True)
                    if not self.multi_publish:
                        raise


class EventPipeline(base.Pipeline):
    """Represents a pipeline for Events."""

    def __str__(self):
        # NOTE(gordc): prepend a namespace so we ensure event and sample
        #              pipelines do not have the same name.
        return 'event:%s' % super().__str__()

    def publish_data(self, events):
        if not isinstance(events, list):
            events = [events]
        supported = [e for e in events if self.supported(e)]
        self.sink.publish_events(supported)

    def supported(self, event):
        return self.source.support_event(event.event_type)


class EventPipelineManager(base.PipelineManager):

    pm_type = 'event'
    pm_pipeline = EventPipeline
    pm_source = EventSource
    pm_sink = EventSink

    def __init__(self, conf):
        super().__init__(
            conf, conf.event_pipeline_cfg_file)

    def get_main_endpoints(self):
        return [EventEndpoint(self.conf, self.publisher())]



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ceilometer import designate_client
from ceilometer.polling import plugin_base


class ZoneDiscovery(plugin_base.DiscoveryBase):
    """Discovery class for Designate DNS zones."""

    KEYSTONE_REQUIRED_FOR_SERVICE = 'designate'

    def __init__(self, conf):
        super().__init__(conf)
        self.designate_client = designate_client.Client(conf)

    def discover(self, manager, param=None):
        """Discover DNS zone resources to monitor."""
        return self.designate_client.zones_list()



# Copyright 2012-2014 eNovance <licensing@enovance.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from oslo_config import cfg
import oslo_i18n
from oslo_log import log
from oslo_reports import guru_meditation_report as gmr
from oslo_reports import opts as gmr_opts

from ceilometer import keystone_client
from ceilometer import messaging
from ceilometer import opts
from ceilometer import sample
from ceilometer import version


def prepare_service(argv=None, config_files=None, conf=None):
    if argv is None:
        argv = sys.argv

    if conf is None:
        conf = cfg.ConfigOpts()

    oslo_i18n.enable_lazy()
    for group, options in opts.list_opts():
        conf.register_opts(list(options),
                           group=None if group == "DEFAULT" else group)
    keystone_client.register_keystoneauth_opts(conf)
    log.register_options(conf)
    log_levels = (conf.default_log_levels +
                  ['futurist=INFO', 'keystoneclient=INFO',
                   'openstack=INFO'])
    log.set_defaults(default_log_levels=log_levels)

    conf(argv[1:], project='ceilometer', validate_default_values=True,
         version=version.version_info.version_string(),
         default_config_files=config_files)

    keystone_client.post_register_keystoneauth_opts(conf)

    log.setup(conf, 'ceilometer')
    sample.setup(conf)

    gmr_opts.set_defaults(conf)
    gmr.TextGuruMeditation.setup_autorun(version, conf=conf)
    messaging.setup()
    return conf



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import fixtures
from openstack.shared_file_system.v2 import share
from oslotest import base

from ceilometer.polling import manager
from ceilometer.polling import plugin_base
from ceilometer import service
from ceilometer.share import discovery
from ceilometer.share import manila


class _BaseTestSharePollster(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.addCleanup(mock.patch.stopall)
        self.CONF = service.prepare_service([], [])
        # Mock the openstack.connection.Connection to avoid auth issues
        with mock.patch('openstack.connection.Connection'):
            self.manager = manager.AgentManager(0, self.CONF)
        plugin_base._get_keystone = mock.Mock()
        catalog = (plugin_base._get_keystone.session.auth.get_access.
                   return_value.service_catalog)
        catalog.get_endpoints = mock.MagicMock(
            return_value={'sharev2': mock.ANY})

    @staticmethod
    def fake_get_shares():
        return [
            share.Share(
                connection=None,
                id='share-1-uuid',
                name='my-share-1',
                availability_zone='az-1',
                share_protocol='NFS',
                share_type='default',
                share_network_id='network-1-uuid',
                status='available',
                host='host-1',
                is_public=False,
                size=100,
                project_id='tenant-1-uuid',
            ),
            share.Share(
                connection=None,
                id='share-2-uuid',
                name='my-share-2',
                availability_zone='az-2',
                share_protocol='CIFS',
                share_type='default',
                share_network_id='network-2-uuid',
                status='creating',
                host='host-2',
                is_public=True,
                size=50,
                project_id='tenant-2-uuid',
            ),
            share.Share(
                connection=None,
                id='share-3-uuid',
                name='my-share-3',
                availability_zone=None,
                share_protocol='NFS',
                share_type='default',
                share_network_id='network-3-uuid',
                status='error',
                host='host-3',
                is_public=False,
                size=200,
                project_id='tenant-1-uuid',
            ),
        ]


class TestShareStatusPollster(_BaseTestSharePollster):

    def setUp(self):
        super().setUp()
        self.pollster = manila.ShareStatusPollster(self.CONF)
        fake_shares = self.fake_get_shares()
        self.useFixture(fixtures.MockPatch(
            'ceilometer.manila_client.Client.shares_list',
            return_value=fake_shares))

    def test_share_get_samples(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_shares()))

        self.assertEqual(len(self.fake_get_shares()), len(samples))
        # Verify metadata fields are correctly extracted
        self.assertEqual('my-share-1', samples[0].resource_metadata['name'])
        self.assertEqual('az-1',
                         samples[0].resource_metadata['availability_zone'])
        # Verify share_protocol is renamed to protocol
        self.assertEqual('NFS', samples[0].resource_metadata['protocol'])
        self.assertNotIn('share_protocol', samples[0].resource_metadata)

    def test_share_status_volume(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_shares()))

        # available = 1, creating = 2, error = 4
        self.assertEqual(1, samples[0].volume)
        self.assertEqual(2, samples[1].volume)
        self.assertEqual(4, samples[2].volume)

    def test_get_share_meter_names(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_shares()))
        self.assertEqual({'manila.share.status'},
                         {s.name for s in samples})

    def test_share_discovery(self):
        with mock.patch('openstack.connection.Connection'):
            discovered_shares = discovery.ShareDiscovery(
                self.CONF).discover(self.manager)
        self.assertEqual(len(self.fake_get_shares()),
                         len(list(discovered_shares)))


class TestShareSizePollster(_BaseTestSharePollster):

    def setUp(self):
        super().setUp()
        self.pollster = manila.ShareSizePollster(self.CONF)

    def test_share_size_volume(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_shares()))
        self.assertEqual(100, samples[0].volume)
        self.assertEqual(50, samples[1].volume)
        self.assertEqual(200, samples[2].volume)

    def test_get_share_meter_names(self):
        samples = list(self.pollster.get_samples(
            self.manager, {},
            resources=self.fake_get_shares()))
        self.assertEqual({'manila.share.size'},
                         {s.name for s in samples})


class TestSharePollsterUnknownStatus(_BaseTestSharePollster):

    def test_unknown_share_status(self):
        pollster = manila.ShareStatusPollster(self.CONF)
        fake_share = share.Share(
            connection=None,
            id='share-unknown-uuid',
            name='my-share-unknown',
            availability_zone=None,
            share_protocol='NFS',
            share_type='default',
            share_network_id='network-unknown-uuid',
            status='UNKNOWN_STATUS',
            host='host-unknown',
            is_public=False,
            size=100,
            project_id='tenant-1-uuid',
        )
        samples = list(pollster.get_samples(
            self.manager, {}, resources=[fake_share]))
        self.assertEqual(1, len(samples))
        self.assertEqual(-1, samples[0].volume)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import datetime
from unittest import mock

import fixtures
from novaclient import exceptions

from ceilometer.compute import discovery
from ceilometer.compute.pollsters import util
from ceilometer import service
from ceilometer.tests import base


# Mock libvirt constants and exceptions to avoid libvirt dependency
VIR_ERR_NO_DOMAIN_METADATA = 80


class FakeLibvirtError(Exception):
    """Fake exception to replace libvirt.libvirtError in tests."""

    def __init__(self, message, error_code=None, error_domain=None):
        super().__init__(message)
        self._error_code = error_code
        self._error_domain = error_domain

    def get_error_code(self):
        return self._error_code

    def get_error_domain(self):
        return self._error_domain


LIBVIRT_METADATA_XML = """
<instance>
  <package version="14.0.0"/>
  <name>test.dom.com</name>
  <creationTime>2016-11-16 07:35:06</creationTime>
  <flavor name="m1.tiny" id="eba4213d-3c6c-4b5f-8158-dd0022d71d62">
    <memory>512</memory>
    <disk>1</disk>
    <swap>0</swap>
    <ephemeral>0</ephemeral>
    <vcpus>1</vcpus>
    <extraSpecs>
      <extraSpec name="hw_rng:allowed">true</extraSpec>
    </extraSpecs>
  </flavor>
  <image uuid="bdaf114a-35e9-4163-accd-226d5944bf11">
    <containerFormat>bare</containerFormat>
    <diskFormat>raw</diskFormat>
    <minDisk>1</minDisk>
    <minRam>0</minRam>
    <properties>
      <property name="os_distro">ubuntu</property>
      <property name="os_type">linux</property>
    </properties>
  </image>
  <owner>
    <user uuid="a1f4684e58bd4c88aefd2ecb0783b497">admin</user>
    <project uuid="d99c829753f64057bc0f2030da309943">admin</project>
  </owner>
  <root type="image" uuid="bdaf114a-35e9-4163-accd-226d5944bf11"/>
</instance>
"""

LIBVIRT_METADATA_XML_OLD = """
<instance>
  <package version="14.0.0"/>
  <name>test.dom.com</name>
  <creationTime>2016-11-16 07:35:06</creationTime>
  <flavor name="m1.tiny">
    <memory>512</memory>
    <disk>1</disk>
    <swap>0</swap>
    <ephemeral>0</ephemeral>
    <vcpus>1</vcpus>
  </flavor>
  <owner>
    <user uuid="a1f4684e58bd4c88aefd2ecb0783b497">admin</user>
    <project uuid="d99c829753f64057bc0f2030da309943">admin</project>
  </owner>
  <root type="image" uuid="bdaf114a-35e9-4163-accd-226d5944bf11"/>
</instance>
"""

LIBVIRT_METADATA_XML_EMPTY_FLAVOR_ID = """
<instance>
  <package version="14.0.0"/>
  <name>test.dom.com</name>
  <creationTime>2016-11-16 07:35:06</creationTime>
  <flavor name="m1.tiny" id="">
    <memory>512</memory>
    <disk>1</disk>
    <swap>0</swap>
    <ephemeral>0</ephemeral>
    <vcpus>1</vcpus>
    <extraSpecs>
      <extraSpec name="hw_rng:allowed">true</extraSpec>
    </extraSpecs>
  </flavor>
  <owner>
    <user uuid="a1f4684e58bd4c88aefd2ecb0783b497">admin</user>
    <project uuid="d99c829753f64057bc0f2030da309943">admin</project>
  </owner>
  <root type="image" uuid="bdaf114a-35e9-4163-accd-226d5944bf11"/>
</instance>
"""

LIBVIRT_METADATA_XML_NO_FLAVOR_ID = """
<instance>
  <package version="14.0.0"/>
  <name>test.dom.com</name>
  <creationTime>2016-11-16 07:35:06</creationTime>
  <flavor name="m1.tiny">
    <memory>512</memory>
    <disk>1</disk>
    <swap>0</swap>
    <ephemeral>0</ephemeral>
    <vcpus>1</vcpus>
    <extraSpecs>
      <extraSpec name="hw_rng:allowed">true</extraSpec>
    </extraSpecs>
  </flavor>
  <owner>
    <user uuid="a1f4684e58bd4c88aefd2ecb0783b497">admin</user>
    <project uuid="d99c829753f64057bc0f2030da309943">admin</project>
  </owner>
  <root type="image" uuid="bdaf114a-35e9-4163-accd-226d5944bf11"/>
</instance>
"""

LIBVIRT_METADATA_XML_EMPTY_FLAVOR_EXTRA_SPECS = """
<instance>
  <package version="14.0.0"/>
  <name>test.dom.com</name>
  <creationTime>2016-11-16 07:35:06</creationTime>
  <flavor name="m1.tiny" id="eba4213d-3c6c-4b5f-8158-dd0022d71d62">
    <memory>512</memory>
    <disk>1</disk>
    <swap>0</swap>
    <ephemeral>0</ephemeral>
    <vcpus>1</vcpus>
    <extraSpecs></extraSpecs>
  </flavor>
  <owner>
    <user uuid="a1f4684e58bd4c88aefd2ecb0783b497">admin</user>
    <project uuid="d99c829753f64057bc0f2030da309943">admin</project>
  </owner>
  <root type="image" uuid="bdaf114a-35e9-4163-accd-226d5944bf11"/>
</instance>
"""

LIBVIRT_METADATA_XML_NO_FLAVOR_EXTRA_SPECS = """
<instance>
  <package version="14.0.0"/>
  <name>test.dom.com</name>
  <creationTime>2016-11-16 07:35:06</creationTime>
  <flavor name="m1.tiny" id="eba4213d-3c6c-4b5f-8158-dd0022d71d62">
    <memory>512</memory>
    <disk>1</disk>
    <swap>0</swap>
    <ephemeral>0</ephemeral>
    <vcpus>1</vcpus>
  </flavor>
  <owner>
    <user uuid="a1f4684e58bd4c88aefd2ecb0783b497">admin</user>
    <project uuid="d99c829753f64057bc0f2030da309943">admin</project>
  </owner>
  <root type="image" uuid="bdaf114a-35e9-4163-accd-226d5944bf11"/>
</instance>
"""

LIBVIRT_METADATA_XML_FROM_VOLUME_IMAGE = """
<instance>
  <package version="14.0.0"/>
  <name>test.dom.com</name>
  <creationTime>2016-11-16 07:35:06</creationTime>
  <flavor name="m1.tiny" id="eba4213d-3c6c-4b5f-8158-dd0022d71d62">
    <memory>512</memory>
    <disk>1</disk>
    <swap>0</swap>
    <ephemeral>0</ephemeral>
    <vcpus>1</vcpus>
    <extraSpecs>
      <extraSpec name="hw_rng:allowed">true</extraSpec>
    </extraSpecs>
  </flavor>
  <image uuid="">
    <containerFormat>bare</containerFormat>
    <diskFormat>raw</diskFormat>
    <minDisk>1</minDisk>
    <minRam>0</minRam>
    <properties>
      <property name="os_distro">ubuntu</property>
      <property name="os_type">linux</property>
    </properties>
  </image>
  <owner>
    <user uuid="a1f4684e58bd4c88aefd2ecb0783b497">admin</user>
    <project uuid="d99c829753f64057bc0f2030da309943">admin</project>
  </owner>
</instance>
"""

LIBVIRT_METADATA_XML_FROM_VOLUME_NO_IMAGE = """
<instance>
  <package version="14.0.0"/>
  <name>test.dom.com</name>
  <creationTime>2016-11-16 07:35:06</creationTime>
  <flavor name="m1.tiny" id="eba4213d-3c6c-4b5f-8158-dd0022d71d62">
    <memory>512</memory>
    <disk>1</disk>
    <swap>0</swap>
    <ephemeral>0</ephemeral>
    <vcpus>1</vcpus>
    <extraSpecs>
      <extraSpec name="hw_rng:allowed">true</extraSpec>
    </extraSpecs>
  </flavor>
  <image uuid="">
    <properties></properties>
  </image>
  <owner>
    <user uuid="a1f4684e58bd4c88aefd2ecb0783b497">admin</user>
    <project uuid="d99c829753f64057bc0f2030da309943">admin</project>
  </owner>
</instance>
"""

LIBVIRT_DESC_XML = """
<domain type='kvm' id='1'>
  <name>instance-00000001</name>
  <uuid>a75c2fa5-6c03-45a8-bbf7-b993cfcdec27</uuid>
  <os>
    <type arch='x86_64' machine='pc-i440fx-xenial'>hvm</type>
    <kernel>/opt/stack/data/nova/instances/a75c2fa5-6c03-45a8-bbf7-b993cfcdec27/kernel</kernel>
    <initrd>/opt/stack/data/nova/instances/a75c2fa5-6c03-45a8-bbf7-b993cfcdec27/ramdisk</initrd>
    <cmdline>root=/dev/vda console=tty0 console=ttyS0</cmdline>
    <boot dev='hd'/>
    <smbios mode='sysinfo'/>
  </os>
</domain>
"""

LIBVIRT_MANUAL_INSTANCE_DESC_XML = """
<domain type='kvm' id='1'>
  <name>Manual-instance-00000001</name>
  <uuid>5e637d0d-8c0e-441a-a11a-a9dc95aed84e</uuid>
  <os>
    <type arch='x86_64' machine='pc-i440fx-xenial'>hvm</type>
    <kernel>/opt/instances/5e637d0d-8c0e-441a-a11a-a9dc95aed84e/kernel</kernel>
    <initrd>/opt/instances/5e637d0d-8c0e-441a-a11a-a9dc95aed84e/ramdisk</initrd>
    <cmdline>root=/dev/vda console=tty0 console=ttyS0</cmdline>
    <boot dev='hd'/>
    <smbios mode='sysinfo'/>
  </os>
</domain>
"""


class FakeDomain:
    def __init__(self, desc=None, metadata=None):
        self._desc = desc or LIBVIRT_DESC_XML
        self._metadata = metadata or LIBVIRT_METADATA_XML

    def state(self):
        return [1, 2]

    def name(self):
        return "instance-00000001"

    def UUIDString(self):
        return "a75c2fa5-6c03-45a8-bbf7-b993cfcdec27"

    def XMLDesc(self):
        return self._desc

    def metadata(self, flags, url):
        return self._metadata


class FakeConn:
    def __init__(self, domains=None):
        self._domains = domains or [FakeDomain()]

    def listAllDomains(self):
        return list(self._domains)

    def isAlive(self):
        return True


class FakeManualInstanceDomain:
    def state(self):
        return [1, 2]

    def name(self):
        return "Manual-instance-00000001"

    def UUIDString(self):
        return "5e637d0d-8c0e-441a-a11a-a9dc95aed84e"

    def XMLDesc(self):
        return LIBVIRT_MANUAL_INSTANCE_DESC_XML

    def metadata(self, flags, url):
        # Note(xiexianbin): vm not create by nova-compute don't have metadata
        # elements like: '<nova:instance
        #  xmlns:nova="http://openstack.org/xmlns/libvirt/nova/1.0">'
        # When invoke get metadata method, raise libvirtError.
        raise FakeLibvirtError(
            "metadata not found: Requested metadata element is not present",
            VIR_ERR_NO_DOMAIN_METADATA)


class FakeManualInstanceConn:
    def listAllDomains(self):
        return [FakeManualInstanceDomain()]

    def isAlive(self):
        return True


class TestDiscovery(base.BaseTestCase):

    def setUp(self):
        super().setUp()

        self.instance = mock.MagicMock()
        self.instance.name = 'instance-00000001'
        setattr(self.instance, 'OS-EXT-SRV-ATTR:instance_name',
                self.instance.name)
        setattr(self.instance, 'OS-EXT-STS:vm_state',
                'active')
        # FIXME(sileht): This is wrong, this should be a uuid
        # The internal id of nova can't be retrieved via API or notification
        self.instance.id = 1
        self.instance.flavor = {'name': 'm1.small',
                                'id': 'eba4213d-3c6c-4b5f-8158-dd0022d71d62',
                                'vcpus': 1,
                                'ram': 512,
                                'disk': 20,
                                'ephemeral': 0,
                                'extra_specs': {'hw_rng:allowed': 'true'}}
        self.instance.status = 'active'
        self.instance.metadata = {
            'fqdn': 'vm_fqdn',
            'metering.stack': '2cadc4b4-8789-123c-b4eg-edd2f0a9c128',
            'project_cos': 'dev'}

        # as we're having lazy hypervisor inspector singleton object in the
        # base compute pollster class, that leads to the fact that we
        # need to mock all this class property to avoid context sharing between
        # the tests
        self.client = mock.MagicMock()
        self.client.instance_get_all_by_host.return_value = [self.instance]
        patch_client = fixtures.MockPatch('ceilometer.nova_client.Client',
                                          return_value=self.client)
        self.useFixture(patch_client)

        self.utc_now = mock.MagicMock(
            return_value=datetime.datetime(
                2016, 1, 1, tzinfo=datetime.timezone.utc))
        patch_timeutils = fixtures.MockPatch('oslo_utils.timeutils.utcnow',
                                             self.utc_now)
        self.useFixture(patch_timeutils)

        self.CONF = service.prepare_service([], [])
        self.CONF.set_override('host', 'test')

        # Add a mocked Libvirt, so we don't have to rely on python-libvirt
        # being installed only for testing
        self.libvirt = mock.MagicMock()
        self.libvirt.libvirtError = FakeLibvirtError
        patch_libvirt = fixtures.MockPatch(
            'ceilometer.compute.virt.libvirt.utils.libvirt', self.libvirt)
        self.useFixture(patch_libvirt)

    def test_normal_discovery(self):
        self.CONF.set_override("instance_discovery_method",
                               "naive",
                               group="compute")
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        self.assertEqual(1, len(resources))
        self.assertEqual(1, list(resources)[0].id)

        self.client.instance_get_all_by_host.assert_called_once_with(
            'test', None)

        resources = dsc.discover(mock.MagicMock())
        self.assertEqual(1, len(resources))
        self.assertEqual(1, list(resources)[0].id)
        self.client.instance_get_all_by_host.assert_called_with(
            self.CONF.host, "2016-01-01T00:00:00+00:00")

    def test_discovery_with_resource_update_interval(self):
        self.CONF.set_override("instance_discovery_method",
                               "naive",
                               group="compute")
        self.CONF.set_override("resource_update_interval", 600,
                               group="compute")
        dsc = discovery.InstanceDiscovery(self.CONF)
        dsc.last_run = datetime.datetime(
            2016, 1, 1, tzinfo=datetime.timezone.utc)

        self.utc_now.return_value = datetime.datetime(
            2016, 1, 1, minute=5, tzinfo=datetime.timezone.utc)
        resources = dsc.discover(mock.MagicMock())
        self.assertEqual(0, len(resources))
        self.client.instance_get_all_by_host.assert_not_called()

        self.utc_now.return_value = datetime.datetime(
            2016, 1, 1, minute=20, tzinfo=datetime.timezone.utc)
        resources = dsc.discover(mock.MagicMock())
        self.assertEqual(1, len(resources))
        self.assertEqual(1, list(resources)[0].id)
        self.client.instance_get_all_by_host.assert_called_once_with(
            self.CONF.host, "2016-01-01T00:00:00+00:00")

    @mock.patch.object(discovery.InstanceDiscovery, "get_server")
    @mock.patch.object(discovery.InstanceDiscovery, "get_flavor_id")
    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt(
            self, mock_libvirt_conn,
            mock_get_flavor_id, mock_get_server):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        mock_libvirt_conn.return_value = FakeConn()
        mock_get_server.return_value = argparse.Namespace(
            metadata={"metering.server_group": "group1"})
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        mock_get_flavor_id.assert_not_called()
        mock_get_server.assert_called_with(
            "a75c2fa5-6c03-45a8-bbf7-b993cfcdec27")

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)
        self.assertEqual("a75c2fa5-6c03-45a8-bbf7-b993cfcdec27",
                         s.resource_id)
        self.assertEqual("d99c829753f64057bc0f2030da309943",
                         s.project_id)
        self.assertEqual("a1f4684e58bd4c88aefd2ecb0783b497",
                         s.user_id)

        metadata = s.resource_metadata
        self.assertEqual(1, metadata["vcpus"])
        self.assertEqual(512, metadata["memory_mb"])
        self.assertEqual(1, metadata["disk_gb"])
        self.assertEqual(0, metadata["ephemeral_gb"])
        self.assertEqual(1, metadata["root_gb"])
        self.assertEqual("bdaf114a-35e9-4163-accd-226d5944bf11",
                         metadata["image_ref"])
        self.assertEqual("test.dom.com", metadata["display_name"])
        self.assertEqual("instance-00000001", metadata["name"])
        self.assertEqual("a75c2fa5-6c03-45a8-bbf7-b993cfcdec27",
                         metadata["instance_id"])
        self.assertEqual("m1.tiny", metadata["instance_type"])
        self.assertEqual({"name": "m1.tiny",
                          "id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62",
                          "ram": 512,
                          "disk": 1,
                          "swap": 0,
                          "ephemeral": 0,
                          "vcpus": 1,
                          "extra_specs": {"hw_rng:allowed": "true"}},
                         metadata["flavor"])
        self.assertEqual(
            "4d0bc931ea7f0513da2efd9acb4cf3a273c64b7bcc544e15c070e662",
            metadata["host"])
        self.assertEqual(self.CONF.host, metadata["instance_host"])
        self.assertEqual("active", metadata["status"])
        self.assertEqual("running", metadata["state"])
        self.assertEqual("hvm", metadata["os_type"])
        self.assertEqual("x86_64", metadata["architecture"])
        self.assertEqual({"server_group": "group1"},
                         metadata["user_metadata"])
        self.assertEqual({"id"},
                         set(metadata["image"].keys()))
        self.assertEqual("bdaf114a-35e9-4163-accd-226d5944bf11",
                         metadata["image"]["id"])
        self.assertIn("image_meta", metadata)
        self.assertEqual({"base_image_ref",
                          "container_format",
                          "disk_format",
                          "min_disk",
                          "min_ram",
                          "os_distro",
                          "os_type"},
                         set(metadata["image_meta"].keys()))
        self.assertEqual("bdaf114a-35e9-4163-accd-226d5944bf11",
                         metadata["image_meta"]["base_image_ref"])
        self.assertEqual("bare",
                         metadata["image_meta"]["container_format"])
        self.assertEqual("raw",
                         metadata["image_meta"]["disk_format"])
        self.assertEqual("1",
                         metadata["image_meta"]["min_disk"])
        self.assertEqual("0",
                         metadata["image_meta"]["min_ram"])
        self.assertEqual("ubuntu",
                         metadata["image_meta"]["os_distro"])
        self.assertEqual("linux",
                         metadata["image_meta"]["os_type"])

    @mock.patch.object(discovery.InstanceDiscovery, "get_server")
    @mock.patch.object(discovery.InstanceDiscovery, "get_flavor_id")
    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_old(
            self, mock_libvirt_conn,
            mock_get_flavor_id, mock_get_server):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        mock_libvirt_conn.return_value = FakeConn(
            domains=[FakeDomain(metadata=LIBVIRT_METADATA_XML_OLD)])
        mock_get_server.return_value = argparse.Namespace(
            flavor={"id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62"},
            metadata={"metering.server_group": "group1"})
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        mock_get_flavor_id.assert_not_called()
        mock_get_server.assert_called_with(
            "a75c2fa5-6c03-45a8-bbf7-b993cfcdec27")

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)
        self.assertEqual("a75c2fa5-6c03-45a8-bbf7-b993cfcdec27",
                         s.resource_id)
        self.assertEqual("d99c829753f64057bc0f2030da309943",
                         s.project_id)
        self.assertEqual("a1f4684e58bd4c88aefd2ecb0783b497",
                         s.user_id)

        metadata = s.resource_metadata
        self.assertEqual(1, metadata["vcpus"])
        self.assertEqual(512, metadata["memory_mb"])
        self.assertEqual(1, metadata["disk_gb"])
        self.assertEqual(0, metadata["ephemeral_gb"])
        self.assertEqual(1, metadata["root_gb"])
        self.assertEqual("bdaf114a-35e9-4163-accd-226d5944bf11",
                         metadata["image_ref"])
        self.assertEqual("test.dom.com", metadata["display_name"])
        self.assertEqual("instance-00000001", metadata["name"])
        self.assertEqual("a75c2fa5-6c03-45a8-bbf7-b993cfcdec27",
                         metadata["instance_id"])
        self.assertEqual("m1.tiny", metadata["instance_type"])
        self.assertEqual({"name": "m1.tiny",
                          "id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62",
                          "ram": 512,
                          "disk": 1,
                          "swap": 0,
                          "ephemeral": 0,
                          "vcpus": 1},
                         metadata["flavor"])
        self.assertEqual(
            "4d0bc931ea7f0513da2efd9acb4cf3a273c64b7bcc544e15c070e662",
            metadata["host"])
        self.assertEqual(self.CONF.host, metadata["instance_host"])
        self.assertEqual("active", metadata["status"])
        self.assertEqual("running", metadata["state"])
        self.assertEqual("hvm", metadata["os_type"])
        self.assertEqual("x86_64", metadata["architecture"])
        self.assertEqual({"server_group": "group1"},
                         metadata["user_metadata"])
        self.assertEqual({"id"},
                         set(metadata["image"].keys()))
        self.assertEqual("bdaf114a-35e9-4163-accd-226d5944bf11",
                         metadata["image"]["id"])
        self.assertNotIn("image_meta", metadata)

    @mock.patch.object(discovery.InstanceDiscovery, "get_server")
    @mock.patch.object(discovery.InstanceDiscovery, "get_flavor_id")
    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_no_extra_metadata(
            self, mock_libvirt_conn,
            mock_get_flavor_id, mock_get_server):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        self.CONF.set_override("fetch_extra_metadata", False, group="compute")
        mock_libvirt_conn.return_value = FakeConn()
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        mock_get_flavor_id.assert_not_called()
        mock_get_server.assert_not_called()

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)

        metadata = s.resource_metadata
        self.assertNotIn("user_metadata", metadata)

    @mock.patch.object(discovery.InstanceDiscovery, "get_server")
    @mock.patch.object(discovery.InstanceDiscovery, "get_flavor_id")
    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_empty_flavor_id_get_by_flavor(
            self, mock_libvirt_conn,
            mock_get_flavor_id, mock_get_server):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        self.CONF.set_override("fetch_extra_metadata", False, group="compute")
        mock_libvirt_conn.return_value = FakeConn(
            domains=[FakeDomain(
                metadata=LIBVIRT_METADATA_XML_EMPTY_FLAVOR_ID)])
        mock_get_flavor_id.return_value = (
            "eba4213d-3c6c-4b5f-8158-dd0022d71d62")
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        mock_get_flavor_id.assert_called_with("m1.tiny")
        mock_get_server.assert_not_called()

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)

        metadata = s.resource_metadata
        self.assertEqual("m1.tiny", metadata["instance_type"])
        self.assertEqual({"name": "m1.tiny",
                          "id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62",
                          "ram": 512,
                          "disk": 1,
                          "swap": 0,
                          "ephemeral": 0,
                          "vcpus": 1,
                          "extra_specs": {"hw_rng:allowed": "true"}},
                         metadata["flavor"])

    @mock.patch.object(discovery.InstanceDiscovery, "get_server")
    @mock.patch.object(discovery.InstanceDiscovery, "get_flavor_id")
    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_empty_flavor_id_get_by_server(
            self, mock_libvirt_conn,
            mock_get_flavor_id, mock_get_server):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        self.CONF.set_override("fetch_extra_metadata", True, group="compute")
        mock_libvirt_conn.return_value = FakeConn(
            domains=[FakeDomain(
                metadata=LIBVIRT_METADATA_XML_EMPTY_FLAVOR_ID)])
        mock_get_server.return_value = argparse.Namespace(
            flavor={"id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62"},
            metadata={})
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        mock_get_flavor_id.assert_not_called()
        mock_get_server.assert_called_with(
            "a75c2fa5-6c03-45a8-bbf7-b993cfcdec27")

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)

        metadata = s.resource_metadata
        self.assertEqual("m1.tiny", metadata["instance_type"])
        self.assertEqual({"name": "m1.tiny",
                          "id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62",
                          "ram": 512,
                          "disk": 1,
                          "swap": 0,
                          "ephemeral": 0,
                          "vcpus": 1,
                          "extra_specs": {"hw_rng:allowed": "true"}},
                         metadata["flavor"])

    @mock.patch.object(discovery.InstanceDiscovery, "get_server")
    @mock.patch.object(discovery.InstanceDiscovery, "get_flavor_id")
    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_no_flavor_id_get_by_flavor(
            self, mock_libvirt_conn,
            mock_get_flavor_id, mock_get_server):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        self.CONF.set_override("fetch_extra_metadata", False, group="compute")
        mock_libvirt_conn.return_value = FakeConn(
            domains=[FakeDomain(metadata=LIBVIRT_METADATA_XML_NO_FLAVOR_ID)])
        mock_get_flavor_id.return_value = (
            "eba4213d-3c6c-4b5f-8158-dd0022d71d62")
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        mock_get_flavor_id.assert_called_with("m1.tiny")
        mock_get_server.assert_not_called()

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)

        metadata = s.resource_metadata
        self.assertEqual("m1.tiny", metadata["instance_type"])
        self.assertEqual({"name": "m1.tiny",
                          "id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62",
                          "ram": 512,
                          "disk": 1,
                          "swap": 0,
                          "ephemeral": 0,
                          "vcpus": 1,
                          "extra_specs": {"hw_rng:allowed": "true"}},
                         metadata["flavor"])

    @mock.patch.object(discovery.InstanceDiscovery, "get_server")
    @mock.patch.object(discovery.InstanceDiscovery, "get_flavor_id")
    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_no_flavor_id_get_by_server(
            self, mock_libvirt_conn,
            mock_get_flavor_id, mock_get_server):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        self.CONF.set_override("fetch_extra_metadata", True, group="compute")
        mock_libvirt_conn.return_value = FakeConn(
            domains=[FakeDomain(metadata=LIBVIRT_METADATA_XML_NO_FLAVOR_ID)])
        mock_get_server.return_value = argparse.Namespace(
            flavor={"id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62"},
            metadata={})
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        mock_get_flavor_id.assert_not_called()
        mock_get_server.assert_called_with(
            "a75c2fa5-6c03-45a8-bbf7-b993cfcdec27")

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)

        metadata = s.resource_metadata
        self.assertEqual("m1.tiny", metadata["instance_type"])
        self.assertEqual({"name": "m1.tiny",
                          "id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62",
                          "ram": 512,
                          "disk": 1,
                          "swap": 0,
                          "ephemeral": 0,
                          "vcpus": 1,
                          "extra_specs": {"hw_rng:allowed": "true"}},
                         metadata["flavor"])

    @mock.patch.object(discovery.InstanceDiscovery, "get_server")
    @mock.patch.object(discovery.InstanceDiscovery, "get_flavor_id")
    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_empty_flavor_extra_specs(
            self, mock_libvirt_conn,
            mock_get_flavor_id, mock_get_server):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        self.CONF.set_override("fetch_extra_metadata", False, group="compute")
        mock_libvirt_conn.return_value = FakeConn(
            domains=[FakeDomain(
                metadata=LIBVIRT_METADATA_XML_EMPTY_FLAVOR_EXTRA_SPECS)])
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        mock_get_flavor_id.assert_not_called()
        mock_get_server.assert_not_called()

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)

        metadata = s.resource_metadata
        self.assertEqual("m1.tiny", metadata["instance_type"])
        self.assertEqual({"name": "m1.tiny",
                          "id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62",
                          "ram": 512,
                          "disk": 1,
                          "swap": 0,
                          "ephemeral": 0,
                          "vcpus": 1,
                          "extra_specs": {}},
                         metadata["flavor"])

    @mock.patch.object(discovery.InstanceDiscovery, "get_server")
    @mock.patch.object(discovery.InstanceDiscovery, "get_flavor_id")
    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_no_flavor_extra_specs(
            self, mock_libvirt_conn,
            mock_get_flavor_id, mock_get_server):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        self.CONF.set_override("fetch_extra_metadata", False, group="compute")
        mock_libvirt_conn.return_value = FakeConn(
            domains=[FakeDomain(
                metadata=LIBVIRT_METADATA_XML_NO_FLAVOR_EXTRA_SPECS)])
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        mock_get_flavor_id.assert_not_called()
        mock_get_server.assert_not_called()

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)

        metadata = s.resource_metadata
        self.assertEqual("m1.tiny", metadata["instance_type"])
        self.assertEqual({"name": "m1.tiny",
                          "id": "eba4213d-3c6c-4b5f-8158-dd0022d71d62",
                          "ram": 512,
                          "disk": 1,
                          "swap": 0,
                          "ephemeral": 0,
                          "vcpus": 1},
                         metadata["flavor"])

    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_from_volume_image(
            self, mock_libvirt_conn):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        self.CONF.set_override("fetch_extra_metadata", False, group="compute")
        mock_libvirt_conn.return_value = FakeConn(
            domains=[
                FakeDomain(metadata=LIBVIRT_METADATA_XML_FROM_VOLUME_IMAGE)])
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)
        self.assertEqual("a75c2fa5-6c03-45a8-bbf7-b993cfcdec27",
                         s.resource_id)
        self.assertEqual("d99c829753f64057bc0f2030da309943",
                         s.project_id)
        self.assertEqual("a1f4684e58bd4c88aefd2ecb0783b497",
                         s.user_id)

        metadata = s.resource_metadata
        self.assertIsNone(metadata["image"])
        self.assertIn("image_meta", metadata)
        self.assertEqual({"base_image_ref",
                          "container_format",
                          "disk_format",
                          "min_disk",
                          "min_ram",
                          "os_distro",
                          "os_type"},
                         set(metadata["image_meta"].keys()))
        self.assertEqual("",
                         metadata["image_meta"]["base_image_ref"])
        self.assertEqual("bare",
                         metadata["image_meta"]["container_format"])
        self.assertEqual("raw",
                         metadata["image_meta"]["disk_format"])
        self.assertEqual("1",
                         metadata["image_meta"]["min_disk"])
        self.assertEqual("0",
                         metadata["image_meta"]["min_ram"])
        self.assertEqual("ubuntu",
                         metadata["image_meta"]["os_distro"])
        self.assertEqual("linux",
                         metadata["image_meta"]["os_type"])

    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_from_volume_no_image(
            self, mock_libvirt_conn):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        self.CONF.set_override("fetch_extra_metadata", False, group="compute")
        mock_libvirt_conn.return_value = FakeConn(
            domains=[
                FakeDomain(
                    metadata=LIBVIRT_METADATA_XML_FROM_VOLUME_NO_IMAGE)])
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())

        self.assertEqual(1, len(resources))
        r = list(resources)[0]
        s = util.make_sample_from_instance(self.CONF, r, "metric", "delta",
                                           "carrot", 1)

        metadata = s.resource_metadata
        self.assertIsNone(metadata["image"])
        self.assertIn("image_meta", metadata)
        self.assertEqual({"base_image_ref"},
                         set(metadata["image_meta"].keys()))
        self.assertEqual("",
                         metadata["image_meta"]["base_image_ref"])

    def test_discovery_with_legacy_resource_cache_cleanup(self):
        self.CONF.set_override("instance_discovery_method", "naive",
                               group="compute")
        self.CONF.set_override("resource_update_interval", 600,
                               group="compute")
        self.CONF.set_override("resource_cache_expiry", 1800,
                               group="compute")
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())
        self.assertEqual(1, len(resources))

        self.utc_now.return_value = datetime.datetime(
            2016, 1, 1, minute=20, tzinfo=datetime.timezone.utc)
        resources = dsc.discover(mock.MagicMock())
        self.assertEqual(1, len(resources))

        self.utc_now.return_value = datetime.datetime(
            2016, 1, 1, minute=31, tzinfo=datetime.timezone.utc)
        resources = dsc.discover(mock.MagicMock())
        self.assertEqual(1, len(resources))

        expected_calls = [mock.call('test', None),
                          mock.call('test', '2016-01-01T00:00:00+00:00'),
                          mock.call('test', None)]
        self.assertEqual(expected_calls,
                         self.client.instance_get_all_by_host.call_args_list)

    @mock.patch("ceilometer.compute.virt.libvirt.utils."
                "refresh_libvirt_connection")
    def test_discovery_with_libvirt_error(self, mock_libvirt_conn):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        mock_libvirt_conn.return_value = FakeManualInstanceConn()
        dsc = discovery.InstanceDiscovery(self.CONF)
        resources = dsc.discover(mock.MagicMock())
        self.assertEqual(0, len(resources))

    def test_get_flavor_id(self):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        fake_flavor = argparse.Namespace(
            id="eba4213d-3c6c-4b5f-8158-dd0022d71d62")
        self.client.nova_client.flavors.find.return_value = fake_flavor
        dsc = discovery.InstanceDiscovery(self.CONF)
        self.assertEqual(fake_flavor.id, dsc.get_flavor_id("m1.tiny"))

    def test_get_flavor_id_notfound(self):
        self.CONF.set_override("instance_discovery_method",
                               "libvirt_metadata",
                               group="compute")
        self.client.nova_client.flavors.find.side_effect = (
            exceptions.NotFound(404))
        dsc = discovery.InstanceDiscovery(self.CONF)
        self.assertIsNone(dsc.get_flavor_id("m1.tiny"))

    def test_get_server(self):
        self.client.nova_client = mock.MagicMock()
        self.client.nova_client.servers = mock.MagicMock()

        fake_server = mock.MagicMock()
        fake_server.metadata = {'metering.server_group': 'group1'}

        fake_flavor = mock.MagicMock()
        fake_flavor.id = 'fake_id'
        fake_server.flavor = fake_flavor

        self.client.nova_client.servers.get = mock.MagicMock(
            return_value=fake_server)
        dsc = discovery.InstanceDiscovery(self.CONF)

        uuid = '123456'

        ret_server = dsc.get_server(uuid)
        self.assertEqual('fake_id', ret_server.flavor.id)
        self.assertEqual({'metering.server_group': 'group1'},
                         ret_server.metadata)

    def test_get_server_notfound(self):
        self.client.nova_client = mock.MagicMock()
        self.client.nova_client.servers = mock.MagicMock()

        self.client.nova_client.servers.get = mock.MagicMock(
            side_effect=exceptions.NotFound(404))
        dsc = discovery.InstanceDiscovery(self.CONF)

        uuid = '123456'

        ret_server = dsc.get_server(uuid)
        self.assertIsNone(ret_server)
